#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests solubility laws for other species"""

import inspect
import logging

import numpy as np
from jaxtyping import ArrayLike

from atmodeller import debug_logger
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.solubility import get_solubility_models
from atmodeller.solubility.core import Solubility
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import unit_conversion

logger: logging.Logger = debug_logger()
logger.setLevel(logging.WARNING)

RTOL: float = 1.0e-8
"""Relative tolerance"""
ATOL: float = 1.0e-8
"""Absolute tolerance"""

LOG10_SHIFT: ArrayLike = 0
IW: RedoxBufferProtocol = IronWustiteBuffer(LOG10_SHIFT)

# Test a non-unity fugacity so the exponent is relevant for a power law solubility.
TEST_FUGACITY: ArrayLike = 2
TEST_TEMPERATURE: ArrayLike = 2000
TEST_PRESSURE: ArrayLike = 2  # bar
# Several models are calibrated in the low GPa range, so use this instead
TEST_PRESSURE_GPA: ArrayLike = 2 * unit_conversion.GPa_to_bar  # GPa
TEST_FO2: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE))
TEST_FO2_GPA: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE_GPA))

logger.info("TEST_FUGACITY = %e bar", TEST_FUGACITY)
logger.info("TEST_TEMPERATURE = %e K", TEST_TEMPERATURE)
logger.info("TEST_PRESSURE = %e bar", TEST_PRESSURE)
logger.info("TEST_PRESSURE_GPA = %e bar", TEST_PRESSURE_GPA)
logger.info("TEST_FO2 = %e bar", TEST_FO2)
logger.info("TEST_FO2_GPA = %e bar", TEST_FO2_GPA)

solubility_models: dict[str, Solubility] = get_solubility_models()


def test_S2_sulfate_andesite_boulliung(check_values) -> None:
    """Tests S as sulfate SO4^2-/S^6+ in andesite :cite:p:`BW22,BW23corr`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_sulfate_andesite_boulliung23"]
    # target_concentration: ArrayLike = 0.0002601378182149385
    target_concentration: ArrayLike = 95
    test_fugacity_S2_sulfate_boulliung_andesite: ArrayLike = 4.16869e-16
    test_fo2_sulfate_boulliung_andesite: ArrayLike = 0.079432823
    test_temperature_sulfate_boulliung_andesite: ArrayLike = 1523
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_S2_sulfate_boulliung_andesite,
        test_temperature_sulfate_boulliung_andesite,
        TEST_PRESSURE,
        test_fo2_sulfate_boulliung_andesite,
    )


def test_S2_sulfide_andesite_boulliung(check_values) -> None:
    """Tests S as sulfide (S^2-) in andesite :cite:p:`BW23`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_sulfide_andesite_boulliung23"]
    # target_concentration: ArrayLike = 2765.534194086474
    target_concentration: ArrayLike = 376
    test_fugacity_S2_sulfide_boulliung_andesite: ArrayLike = 0.089125094
    test_fo2_sulfide_boulliung_andesite: ArrayLike = 3.98107e-10
    test_temperature_sulfide_boulliung_andesite: ArrayLike = 1473
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_S2_sulfide_boulliung_andesite,
        test_temperature_sulfide_boulliung_andesite,
        TEST_PRESSURE,
        test_fo2_sulfide_boulliung_andesite,
    )


def test_S2_andesite_boulliung(check_values) -> None:
    """Tests S in andesite accounting for both sulfide and sulfate :cite:p:`BW22,BW23corr,BW23`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_andesite_boulliung23"]
    target_concentration: ArrayLike = 2765.5344542242924

    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        TEST_FUGACITY,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_S2_sulfate_basalt_boulliung(check_values) -> None:
    """Tests S in basalt as sulfate, SO4^2-/S^6+ :cite:p:`BW22,BW23corr`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_sulfate_basalt_boulliung23"]
    # target_concentration: ArrayLike = 0.00020528360011091678
    target_concentration: ArrayLike = 210
    test_fugacity_S2_sulfate_boulliung_basalt: ArrayLike = 1.38038e-14
    test_fo2_sulfate_boulliung_basalt: ArrayLike = 0.079432823
    test_temperature_sulfate_boulliung_basalt: ArrayLike = 1623
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_S2_sulfate_boulliung_basalt,
        test_temperature_sulfate_boulliung_basalt,
        TEST_PRESSURE,
        test_fo2_sulfate_boulliung_basalt,
    )


def test_S2_sulfide_basalt_boulliung(check_values) -> None:
    """Tests S in basalt as sulfide (S^2-) :cite:p:`BW23`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_sulfide_basalt_boulliung23"]
    # target_concentration: ArrayLike = 7576.212242182479
    target_concentration: ArrayLike = 760
    test_fugacity_S2_sulfide_boulliung_basalt: ArrayLike = 0.089125094
    test_fo2_sulfide_boulliung_basalt: ArrayLike = 3.98107e-10
    test_temperature_sulfide_boulliung_basalt: ArrayLike = 1473
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_S2_sulfide_boulliung_basalt,
        test_temperature_sulfide_boulliung_basalt,
        TEST_PRESSURE,
        test_fo2_sulfide_boulliung_basalt,
    )


def test_S2_basalt_boulliung(check_values) -> None:
    """Tests S in basalt due to sulfide and sulfate :cite:p:`BW22,BW23corr,BW23`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_basalt_boulliung23"]
    target_concentration: ArrayLike = 7576.212447466079

    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        TEST_FUGACITY,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_S2_sulfate_trachybasalt_boulliung(check_values) -> None:
    """Tests S as sulfate SO4^2-/S^6+ in trachybasalt :cite:p:`BW22,BW23corr`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_sulfate_trachybasalt_boulliung23"]
    # target_concentration: ArrayLike = 0.0007002721319248917
    target_concentration: ArrayLike = 230
    test_fugacity_S2_sulfate_boulliung_trachybasalt: ArrayLike = 2.51189e-15
    test_fo2_sulfate_boulliung_trachybasalt: ArrayLike = 0.079432823
    test_temperature_sulfate_boulliung_trachybasalt: ArrayLike = 1573
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_S2_sulfate_boulliung_trachybasalt,
        test_temperature_sulfate_boulliung_trachybasalt,
        TEST_PRESSURE,
        test_fo2_sulfate_boulliung_trachybasalt,
    )


def test_S2_sulfide_trachybasalt_boulliung(check_values) -> None:
    """Tests S as sulfide (S^2-) in trachybasalt :cite:p:`BW23`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_sulfide_trachybasalt_boulliung23"]
    # target_concentration: ArrayLike = 9573.602305681254
    target_concentration: ArrayLike = 1157
    test_fugacity_S2_sulfide_boulliung_trachybasalt: ArrayLike = 0.089125094
    test_fo2_sulfide_boulliung_trachybasalt: ArrayLike = 3.98107e-10
    test_temperature_sulfide_boulliung_trachybasalt: ArrayLike = 1473
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_S2_sulfide_boulliung_trachybasalt,
        test_temperature_sulfide_boulliung_trachybasalt,
        TEST_PRESSURE,
        test_fo2_sulfide_boulliung_trachybasalt,
    )


def test_S2_trachybasalt_boulliung(check_values) -> None:
    """Tests S in trachybasalt by sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["S2_trachybasalt_boulliung23"]
    target_concentration: ArrayLike = 9573.603005953386

    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        TEST_FUGACITY,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests solubility laws for hydrogen species"""

import inspect
import logging

import numpy as np
from jaxtyping import ArrayLike

from atmodeller import debug_logger
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.solubility import get_solubility_models
from atmodeller.solubility.core import Solubility
from atmodeller.thermodata import IronWustiteBuffer

logger: logging.Logger = debug_logger()
logger.setLevel(logging.WARNING)

# RTOL: float = 1.0e-8
RTOL: float = 0.05
"""Relative tolerance"""
# ATOL: float = 1.0e-8
ATOL: float = 0.05
"""Absolute tolerance"""

LOG10_SHIFT: ArrayLike = 0
IW: RedoxBufferProtocol = IronWustiteBuffer(LOG10_SHIFT)

# Test a non-unity fugacity so the exponent is relevant for a power law solubility.
TEST_FUGACITY: ArrayLike = 2  # bar
TEST_TEMPERATURE: ArrayLike = 2000  # K
TEST_PRESSURE: ArrayLike = 10  # bar
TEST_FO2: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE))

logger.info("TEST_FUGACITY = %e bar", TEST_FUGACITY)
logger.info("TEST_TEMPERATURE = %e K", TEST_TEMPERATURE)
logger.info("TEST_PRESSURE = %e bar", TEST_PRESSURE)
logger.info("TEST_FO2 = %e bar", TEST_FO2)

solubility_models: dict[str, Solubility] = get_solubility_models()


def test_H2_andesite_hirschmann(check_values) -> None:
    """Tests H2 in synthetic andesite :cite:p:`HWA12`.

    Reference Parameters (fH2, H2 Conc) from Table 2 Values for Andesite, Experiment 901
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2_andesite_hirschmann12"]
    target_concentration: ArrayLike = 9000
    test_fugacity_H2_hirschmann_andesite: ArrayLike = 72269
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2_hirschmann_andesite,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2_basalt_hirschmann(check_values) -> None:
    """Tests H2 in synthetic basalt :cite:p:`HWA12`.

    Reference Parameters (fH2, H2 Conc) from Table 2 Values for Basalt, Average of Experiments A697
    and A711.
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2_basalt_hirschmann12"]
    target_concentration: ArrayLike = 2200
    test_fugacity_H2_hirschmann_basalt: ArrayLike = 19058
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2_hirschmann_basalt,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2_silicic_melts_gaillard(check_values) -> None:
    """Tests Fe-H redox exchange in silicate glasses :cite:p:`GSM03`.

    Reference Parameters (fH2 and H2 Conc) from Table 4, No. 29
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2_silicic_melts_gaillard03"]
    target_concentration: ArrayLike = 2.1
    test_fugacity_H2_gaillard: ArrayLike = 8
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2_gaillard,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2O_ano_dio_newcombe(check_values) -> None:
    """Tests H2O in anorthite-diopside-eutectic compositions :cite:p:`NBB17`.

    Reference Parameters (fH2O and H2O Conc) from Table 2, Experiment AD26
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2O_ano_dio_newcombe17"]
    target_concentration: ArrayLike = 229
    test_fugacity_H2O_newcombe_andio: ArrayLike = 0.1
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2O_newcombe_andio,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2O_basalt_dixon(check_values) -> None:
    """Tests H2O in MORB liquids :cite:p:`DSH95`.

    Reference Parameters (fH2O and H2O Conc) from Table 5, fH2O=25 bar
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2O_basalt_dixon95"]
    target_concentration: ArrayLike = 5200
    test_fugacity_H2O_dixon_basalt: ArrayLike = 25
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2O_dixon_basalt,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2O_basalt_mitchell(check_values) -> None:
    """Tests H2O in basaltic melt :cite:p:`MGO17`.

    Reference Parameters (fH2O and H2O Conc) from Figure 7, Maroon Point from 'This Study'.
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2O_basalt_mitchell17"]
    target_concentration: ArrayLike = 202247
    test_fugacity_H2O_mitchell_basalt: ArrayLike = 20600.63
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2O_mitchell_basalt,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2O_lunar_glass_newcombe(check_values) -> None:
    """Tests H2O in lunar basalt :cite:p:`NBB17`.

    Reference Parameters (fH2O and H2O Conc) from Table 2, Average of Experiments LG2 and LG4
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2O_lunar_glass_newcombe17"]
    target_concentration: ArrayLike = 353.5
    test_fugacity_H2O_newcombe_lunarglass: ArrayLike = 0.27
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2O_newcombe_lunarglass,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_H2O_peridotite_sossi(check_values) -> None:
    """Tests H2O in peridotite liquids :cite:p:`STB23`.

    Reference Parameters (fH2O and H2O Conc) from Table 1, Sample Per-7, using epsilon_3550 of 5.1.
    """

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["H2O_peridotite_sossi23"]
    target_concentration: ArrayLike = 42.9
    test_fugacity_H2O_sossi_peridotite: ArrayLike = 0.0038
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_H2O_sossi_peridotite,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )



'''
Collin Cherubim
October 3, 2022
Miscellaneous functions for planetary orbital parameters
'''

import numpy as np
from constants import *

def Luminosity(R, T):
    '''
    Input: R, stellar radius [m]; T, stellar temp [K]
    Output: luminosity [W]
    '''
    L = 4*np.pi*R**2*sbc*T**4
    return L

def SemiMajor(M, P):
    '''
    Input: M, stellar mass [kg]; P, orbital period [s]
    Output: semi-major axis [m]
    '''
    a = ((G*M/4/np.pi**2)*P**2)**(1/3)
    return a

def Period(a, M):
    '''
    Input: M, stellar mass [kg]; a, semi-major axis [m]
    Output: orbital period [s]
    '''
    P = np.sqrt(a**3*4*np.pi**2/G/M)
    return P

def Insolation(L, a):
    '''
    Input: L, stellar luminosity [W]; a, semi-major axis [m]
    Output: incident stellar flux [W/m2]
    '''
    I = L/(4*np.pi*a**2)
    return I

def EqTemp(I, A = 0):
    '''
    Input: planetary incident bolometric flux [W/m2]
    Output: eq temp [K]
    '''
    T = (I*(1 - A)/4/sbc)**(1/4)
    return T

def HabZone(L, Teff, HZ = 'runaway greenhouse'):
    '''
    Calculates HZ distance based on Kopparapu et al 2013
    Inputs:
        - L: stellar luminosity [W]
        - Teff: stellar effective temperature [K]
        - HZ: desired habitable zone
            - 'recent venus'
            - 'runaway greenhouse' (default)
            - 'moist greenhouse'
            - 'max greenhouse'
            - 'early mars'
    Output: orbital distance or corresponding HZ limit [m]
    '''
    if HZ == 'recent venus':
        Seff_solar = 1.7763
        a = 1.4335e-4
        b = 3.3954e-9
        c = -7.6364e-12
        d = -1.1950e-15
    elif HZ == 'runaway greenhouse':
        Seff_solar = 1.0385
        a = 1.2456e-4
        b = 1.4612e-8
        c = -7.6345e-12
        d = -1.7511e-15
    elif HZ == 'moist greenhouse':
        Seff_solar = 1.0146
        a = 8.1884e-5 
        b = 1.9394e-9
        c = -4.3618e-12 
        d = -6.8260e-16
    elif HZ == 'max greenhouse':
        Seff_solar = 0.3507
        a = 5.9578e-5 
        b = 1.6707e-9
        c = -3.0058e-12 
        d = -5.1925e-16
    elif HZ == 'early mars':
        Seff_solar = 0.3207
        a = 5.4471e-5 
        b = 1.5275e-9
        c = -2.1709e-12 
        d = -3.8282e-16

    T_star = Teff - 5780 #[K]
    Seff = Seff_solar + a*T_star + b*T_star**2 + c*T_star**3 + d*T_star**4 # HZ stellar flux
    d = (L/Ls/Seff)**0.5*au2m # orbtial distance of corresponding HZ
    return d



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests solubility laws for carbon species"""

import inspect
import logging

import numpy as np
from jaxtyping import ArrayLike

from atmodeller import debug_logger
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.solubility import get_solubility_models
from atmodeller.solubility.core import Solubility
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import unit_conversion

logger: logging.Logger = debug_logger()
logger.setLevel(logging.WARNING)

RTOL: float = 1e-8
"""Relative tolerance"""
ATOL: float = 1e-8
"""Absolute tolerance"""

LOG10_SHIFT: ArrayLike = 0
IW: RedoxBufferProtocol = IronWustiteBuffer(LOG10_SHIFT)

# Test a non-unity fugacity so the exponent is relevant for a power law solubility.
TEST_FUGACITY: ArrayLike = 2  # bar
TEST_TEMPERATURE: ArrayLike = 2000  # K
TEST_PRESSURE: ArrayLike = 500  # bar, motivated by Dixon experimental range
TEST_FO2: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE))
# Several models are calibrated in the low GPa range, so use this instead for testing
TEST_PRESSURE_GPA: ArrayLike = 2 * unit_conversion.GPa_to_bar  # GPa
TEST_FO2_GPA: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE_GPA))

logger.info("TEST_FUGACITY = %e bar", TEST_FUGACITY)
logger.info("TEST_TEMPERATURE = %e K", TEST_TEMPERATURE)
logger.info("TEST_PRESSURE = %e bar", TEST_PRESSURE)
logger.info("TEST_PRESSURE_GPA = %e bar", TEST_PRESSURE_GPA)
logger.info("TEST_FO2 = %e bar", TEST_FO2)
logger.info("TEST_FO2_GPA = %e bar", TEST_FO2_GPA)

solubility_models: dict[str, Solubility] = get_solubility_models()


def test_CH4_basalt_ardia(check_values) -> None:
    """Tests CH4 in haplobasalt (Fe-free) silicate melt :cite:p:`AHW13`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["CH4_basalt_ardia13"]
    # target_concentration: ArrayLike = 0.0005831884445042942
    target_concentration: ArrayLike = 69.39
    test_fugacity_CH4_ardia_basalt: ArrayLike = 19360
    test_pressureGPa_CH4_ardia_basalt: ArrayLike = 7000
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_CH4_ardia_basalt,
        TEST_TEMPERATURE,
        test_pressureGPa_CH4_ardia_basalt,
        TEST_FO2_GPA,
    )


def test_CO_basalt_armstrong(check_values) -> None:
    """Tests volatiles in mafic melts under reduced conditions :cite:p:`AHS15`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["CO_basalt_armstrong15"]
    # target_concentration: ArrayLike = 0.027396953726422667
    target_concentration: ArrayLike = 17.3
    test_fugacity_CO_armstrong_basalt: ArrayLike = 1000
    test_pressure_CO_armstrong_basalt: ArrayLike = 1.20e4
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_CO_armstrong_basalt,
        TEST_TEMPERATURE,
        test_pressure_CO_armstrong_basalt,
        TEST_FO2_GPA,
    )


def test_CO_basalt_yoshioka(check_values) -> None:
    """Tests carbon in silicate melts :cite:p:`YNN19`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["CO_basalt_yoshioka19"]
    # target_concentration: ArrayLike = 0.1098560543306116
    target_concentration: ArrayLike = 59
    test_fugacity_CO_yoshioka_basalt: ArrayLike = 5248
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_CO_yoshioka_basalt,
        TEST_TEMPERATURE,
        TEST_PRESSURE_GPA,
        TEST_FO2_GPA,
    )


def test_CO_rhyolite_yoshioka(check_values) -> None:
    """Tests carbon in silicate melts :cite:p:`YNN19`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["CO_rhyolite_yoshioka19"]
    # target_concentration: ArrayLike = 1.19271202468211
    target_concentration: ArrayLike = 161
    test_fugacity_CO_yoshioka_rhyolite: ArrayLike = 31623
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_CO_yoshioka_rhyolite,
        TEST_TEMPERATURE,
        TEST_PRESSURE_GPA,
        TEST_FO2_GPA,
    )


def test_CO2_basalt_dixon(check_values) -> None:
    """Tests CO2 in MORB liquids :cite:p:`DSH95`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["CO2_basalt_dixon95"]
    # target_concentration: ArrayLike = 0.8527333099685608
    target_concentration: ArrayLike = 11
    test_fugacity_CO2_dixon_basalt: ArrayLike = 25
    test_temperature_CO2_dixon_basalt: ArrayLike = 1473.15
    test_pressure_CO2_dixon_basalt: ArrayLike = 25
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_CO2_dixon_basalt,
        test_temperature_CO2_dixon_basalt,
        test_pressure_CO2_dixon_basalt,
        TEST_FO2,
    )



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Core classes and functions for thermochemical and critical data"""

import importlib.resources
from contextlib import AbstractContextManager
from dataclasses import dataclass
from importlib.abc import Traversable
from pathlib import Path
from typing import cast

import equinox as eqx
import jax.numpy as jnp
import pandas as pd
from jaxtyping import Array, ArrayLike, Bool, Float, Integer
from molmass import Formula
from xmmutablemap import ImmutableMap

from atmodeller import TEMPERATURE_REFERENCE
from atmodeller.constants import GAS_CONSTANT
from atmodeller.utilities import as_j64, to_native_floats, unit_conversion

DATA_DIRECTORY: Traversable = importlib.resources.files(f"{__package__}.data")
"""Data directory"""
THERMODYNAMIC_DATA_SOURCE: Path = Path("nasa_glenn_coefficients.txt")
"""Source of the thermodynamic data"""
CRITICAL_DATA_SOURCE: Path = Path("critical_data.txt")
"""Source of the critical data"""


class CondensateActivity(eqx.Module):
    """Activity of a stable condensate

    Args:
        activity: Activity. Defaults to 1.
    """

    activity: float = eqx.field(converter=float, default=1)
    """Activity"""

    def log_activity(self, temperature: ArrayLike, pressure: ArrayLike) -> Float[Array, ""]:
        """Log activity

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Log activity, which is dimensionless
        """
        del temperature
        del pressure

        return jnp.log(self.activity)


class ThermodynamicCoefficients(eqx.Module):
    """NASA Glenn coefficients for the thermodynamic properties of an individual species

    Coefficients are available at https://ntrs.nasa.gov/citations/20020085330

    Args:
        b1: Enthalpy constant(s) of integration
        b2: Entropy constant(s) of integration
        cp_coeffs: Heat capacity coefficients
        T_min: Minimum temperature(s) in K in the range
        T_max: Maximum temperature(s) in K in the range
    """

    b1: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """Enthalpy constant(s) of integration"""
    b2: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """Entropy constant(s) of integration"""
    cp_coeffs: tuple[tuple[float, ...], ...] = eqx.field(converter=to_native_floats)
    """Heat capacity coefficients"""
    T_min: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """Minimum temperature(s) in K in the range"""
    T_max: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """Maximum temperature(s) in K in the range"""

    def _get_index(self, temperature: ArrayLike) -> Integer[Array, " T"]:
        """Gets the index of the temperature range for the given temperature

        This assumes the temperature is within one of the ranges and will produce unexpected output
        if the temperature is outside the ranges.

        Args:
            temperature: Temperature in K

        Returns:
            Index of the temperature range
        """
        temperature = jnp.atleast_1d(as_j64(temperature))
        T_max: Array = as_j64(self.T_max)
        T_min: Array = as_j64(self.T_min)

        # Reshape for broadcasting
        bool_mask: Bool[Array, "N T"] = (T_min[:, None] <= temperature[None, :]) & (
            temperature[None, :] <= T_max[:, None]
        )
        index: Integer[Array, " T"] = jnp.argmax(bool_mask, axis=0)

        return index

    def _cp_over_R(
        self, cp_coefficients: Float[Array, "T 7"], temperature: ArrayLike
    ) -> Float[Array, " T"]:
        """Heat capacity relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`

        Args:
            cp_coefficients: Heat capacity coefficients
            temperature: Temperature in K

        Returns:
            Heat capacity relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
        """
        temperature = jnp.atleast_1d(as_j64(temperature))
        temperature_terms: Float[Array, "T 7"] = jnp.stack(
            [
                jnp.power(temperature, -2),
                jnp.power(temperature, -1),
                jnp.ones_like(temperature),
                temperature,
                jnp.power(temperature, 2),
                jnp.power(temperature, 3),
                jnp.power(temperature, 4),
            ],
            axis=-1,
        )

        heat_capacity: Float[Array, " T"] = jnp.einsum(
            "ti,ti->t", cp_coefficients, temperature_terms
        )

        return heat_capacity

    def _S_over_R(
        self, cp_coefficients: Float[Array, "T 7"], b2: ArrayLike, temperature: ArrayLike
    ) -> Float[Array, " T"]:
        """Entropy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`

        Args:
            cp_coefficients: Heat capacity coefficients
            b2: Entropy integration constant
            temperature: Temperature in K

        Returns:
            Entropy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
        """
        temperature = jnp.atleast_1d(as_j64(temperature))
        temperature_terms: Float[Array, "T 7"] = jnp.stack(
            [
                -jnp.power(temperature, -2) / 2,
                -jnp.power(temperature, -1),
                jnp.log(temperature),
                temperature,
                jnp.power(temperature, 2) / 2,
                jnp.power(temperature, 3) / 3,
                jnp.power(temperature, 4) / 4,
            ],
            axis=-1,
        )

        entropy: Float[Array, " T"] = (
            jnp.einsum("ti,ti->t", cp_coefficients, temperature_terms) + b2
        )

        return entropy

    def _H_over_RT(
        self, cp_coefficients: Float[Array, "T 7"], b1: ArrayLike, temperature: ArrayLike
    ) -> Float[Array, " T"]:
        r"""Enthalpy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
        :math:`\times T`

        Args:
            cp_coefficients: Heat capacity coefficients as an array
            b1: Enthalpy integration constant
            temperature: Temperature in K

        Returns:
            Enthalpy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
            :math:`\times T`
        """
        temperature = jnp.atleast_1d(as_j64(temperature))
        temperature_terms: Float[Array, "T 7"] = jnp.stack(
            [
                -jnp.power(temperature, -2),
                jnp.log(temperature) / temperature,
                jnp.ones_like(temperature),
                temperature / 2,
                jnp.power(temperature, 2) / 3,
                jnp.power(temperature, 3) / 4,
                jnp.power(temperature, 4) / 5,
            ],
            axis=-1,
        )

        enthalpy: Float[Array, " T"] = (
            jnp.einsum("ti,ti->t", cp_coefficients, temperature_terms) + b1 / temperature
        )

        return enthalpy

    def _G_over_RT(
        self,
        cp_coefficients: Float[Array, "T 7"],
        b1: ArrayLike,
        b2: ArrayLike,
        temperature: ArrayLike,
    ) -> Float[Array, " T"]:
        r"""Gibbs energy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
        :math:`\times T`

        Args:
            cp_coefficients: Heat capacity coefficients as an array
            b1: Enthalpy integration constant
            b2: Entropy integration constant
            temperature: Temperature in K

        Returns:
            Gibbs energy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
            :math:`\times T`
        """
        enthalpy: Float[Array, " T"] = self._H_over_RT(cp_coefficients, b1, temperature)
        # jax.debug.print("enthalpy = {out}", out=enthalpy)
        entropy: Float[Array, " T"] = self._S_over_R(cp_coefficients, b2, temperature)
        # jax.debug.print("entropy = {out}", out=entropy)
        # No temperature multiplication is correct since the return is Gibbs energy relative to RT
        gibbs: Float[Array, " T"] = enthalpy - entropy

        return gibbs

    def get_gibbs_over_RT(self, temperature: ArrayLike) -> Float[Array, " T"]:
        r"""Gets Gibbs energy to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
        :math:`\times T`

        Args:
            temperature: Temperature in K

        Returns:
            Gibbs energy relative to :data:`GAS_CONSTANT <atmodeller.constants.GAS_CONSTANT>`
            :math:`\times T`
        """
        index: Integer[Array, " T"] = self._get_index(temperature)
        # jax.debug.print("index = {out}", out=index)
        cp_coeffs_for_index: Float[Array, "T 7"] = jnp.take(
            jnp.array(self.cp_coeffs), index, axis=0
        )
        # jax.debug.print("cp_coeffs_for_index = {out}", out=cp_coeffs_for_index)
        b1_for_index: Float[Array, " T"] = jnp.take(jnp.array(self.b1), index)
        # jax.debug.print("b1_for_index = {out}", out=b1_for_index)
        b2_for_index: Float[Array, " T"] = jnp.take(jnp.array(self.b2), index)
        # jax.debug.print("b2_for_index = {out}", out=b2_for_index)
        gibbs_for_index: Float[Array, " T"] = self._G_over_RT(
            cp_coeffs_for_index, b1_for_index, b2_for_index, temperature
        )

        return gibbs_for_index

    def cp(self, temperature: ArrayLike) -> Float[Array, " T"]:
        r"""Gets heat capacity.

        This is :math:`C_p^\circ` in the JANAF tables.

        Args:
            temperature: Temperature in K

        Returns:
            Heat capacity in :math:`\mathrm{J}\ \mathrm{K}^{-1} \mathrm{mol}^{-1}`
        """
        index: Integer[Array, " T"] = self._get_index(temperature)
        cp_coeffs_for_index: Float[Array, "T 7"] = jnp.take(
            jnp.array(self.cp_coeffs), index, axis=0
        )
        # jax.debug.print("cp_coeffs_for_index = {out}", out=cp_coeffs_for_index.shape)
        cp: Float[Array, " T"] = self._cp_over_R(cp_coeffs_for_index, temperature) * GAS_CONSTANT

        return cp

    def enthalpy(self, temperature: ArrayLike) -> Float[Array, " T"]:
        r"""Gets enthalpy.

        This is :math:`H` in the JANAF tables.

        Args:
            temperature: Temperature in K

        Returns:
            Enthalpy in :math:`\mathrm{J}\ \mathrm{mol}^{-1}`
        """
        index: Integer[Array, " T"] = self._get_index(temperature)
        cp_coeffs_for_index: Float[Array, "T 7"] = jnp.take(
            jnp.array(self.cp_coeffs), index, axis=0
        )
        b1_for_index: Float[Array, " T"] = jnp.take(jnp.array(self.b1), index)
        enthalpy: Float[Array, " T"] = (
            self._H_over_RT(cp_coeffs_for_index, b1_for_index, temperature)
            * GAS_CONSTANT
            * temperature
        )

        return enthalpy

    def reference_enthalpy(self) -> Float[Array, ""]:
        r"""Gets reference enthalpy.

        This is :math:`H^{\circ}(T_r)` in the JANAF tables.

        Args:
            temperature: Temperature in K

        Returns:
            Reference enthalpy in :math:`\mathrm{J}\ \mathrm{mol}^{-1}`
        """
        index: Integer[Array, ""] = self._get_index(TEMPERATURE_REFERENCE)
        # jax.debug.print("index = {out}", out=index)
        cp_coeffs_for_index: Float[Array, "7"] = jnp.take(jnp.array(self.cp_coeffs), index, axis=0)
        b1_for_index: Float[Array, ""] = jnp.take(jnp.array(self.b1), index)
        # jax.debug.print("b1_for_index = {out}", out=b1_for_index)
        reference_enthalpy: Float[Array, ""] = (
            self._H_over_RT(cp_coeffs_for_index, b1_for_index, TEMPERATURE_REFERENCE)
            * GAS_CONSTANT
            * TEMPERATURE_REFERENCE
        )

        return reference_enthalpy

    def enthalpy_function(self, temperature: ArrayLike) -> Float[Array, " T"]:
        r"""Gets enthalpy function/increment.

        This is :math:`H-H^{\circ}(T_r)` in the JANAF tables.

        Args:
            temperature: Temperature in K

        Returns:
            Enthalpy increment in :math:`\mathrm{J}\ \mathrm{mol}^{-1}`
        """
        enthalpy: Float[Array, " T"] = self.enthalpy(temperature)
        reference_enthalpy: Float[Array, ""] = self.reference_enthalpy()

        return enthalpy - reference_enthalpy

    def entropy(self, temperature: ArrayLike) -> Float[Array, " T"]:
        r"""Gets entropy

        This is :math:`S^\circ` in the JANAF tables.

        Args:
            temperature: Temperature in K

        Returns:
            Entropy in :math:`\mathrm{J}\ \mathrm{K}^{-1} \mathrm{mol}^{-1}`
        """
        index: Integer[Array, " T"] = self._get_index(temperature)
        cp_coeffs_for_index: Float[Array, "T 7"] = jnp.take(
            jnp.array(self.cp_coeffs), index, axis=0
        )
        b2_for_index: Float[Array, " T"] = jnp.take(jnp.array(self.b2), index)
        entropy: Float[Array, " T"] = (
            self._S_over_R(cp_coeffs_for_index, b2_for_index, temperature) * GAS_CONSTANT
        )

        return entropy

    def gibbs_function(self, temperature: ArrayLike) -> Float[Array, " T"]:
        r"""Gets Gibbs energy function.

        This is :math:`-[G^\circ-H^{\circ}(T_r)]/T` in the JANAF tables.

        Args:
            temperature: Temperature in K

        Returns:
            Gibbs energy function in :math:`\mathrm{J}\ \mathrm{K}^{-1} \mathrm{mol}^{-1}`
        """
        gibbs: Float[Array, " T"] = (
            self.get_gibbs_over_RT(temperature) * GAS_CONSTANT * temperature
        )
        gibbs_function: Float[Array, " T"] = -(gibbs - self.reference_enthalpy()) / temperature

        return gibbs_function


@dataclass
class ThermodynamicDataSource:
    """Thermodynamic data source for all species"""

    data: pd.DataFrame
    """Thermodynamic data for all species"""

    def __init__(self):
        data: AbstractContextManager[Path] = importlib.resources.as_file(
            DATA_DIRECTORY.joinpath(THERMODYNAMIC_DATA_SOURCE)  # type: ignore
        )
        with data as datapath:
            self.data = pd.read_csv(datapath, sep=" ", comment="#")

    @property
    def formula_column(self) -> str:
        """Name of the column that refers to the hill formula"""
        return "hill_formula"

    @property
    def state_column(self) -> str:
        """Name of the column that refers to the state of aggregation"""
        return "state"

    def available_species(self) -> tuple[str, ...]:
        """Available species

        Returns:
            Available species
        """
        df: pd.DataFrame = cast(
            pd.DataFrame, self.data[[self.formula_column, self.state_column]].drop_duplicates()
        )
        available_species: tuple[str, ...] = tuple(
            f"{getattr(row, self.formula_column)}_{getattr(row, self.state_column)}"
            for row in df.itertuples(index=False)
        )

        return available_species

    def create_dictionary(self) -> dict[str, ThermodynamicCoefficients]:
        """Dictionary of thermodynamic coefficients for all species

        Returns:
            Dictionary of thermodynamic coefficients for all species
        """
        unique_combinations: pd.DataFrame = cast(
            pd.DataFrame, self.data[[self.formula_column, self.state_column]].drop_duplicates()
        )
        coefficient_dict: dict[str, ThermodynamicCoefficients] = {}

        for row in unique_combinations.itertuples(index=False):
            hill_formula: str = str(getattr(row, self.formula_column))
            state: str = str(getattr(row, self.state_column))
            name: str = f"{hill_formula}_{state}"

            # Find all data across all temperature ranges
            df: pd.DataFrame = cast(
                pd.DataFrame,
                self.data[
                    (self.data[self.formula_column] == hill_formula)
                    & (self.data[self.state_column] == state)
                ],
            )
            cp_coeffs: pd.DataFrame | pd.Series = df[["a1", "a2", "a3", "a4", "a5", "a6", "a7"]]
            coefficient_dict[name] = ThermodynamicCoefficients(
                df["b1"], df["b2"], cp_coeffs, df["T_min"], df["T_max"]
            )

        return coefficient_dict


class CriticalData(eqx.Module):
    """Critical temperature and pressure of a gas species

    Args:
        temperature: Critical temperature in K
        pressure: Critical pressure in bar
    """

    temperature: float = eqx.field(converter=float, default=1)
    """Critical temperature in K"""
    pressure: float = eqx.field(converter=float, default=1)
    """Critical pressure in bar"""


@dataclass
class CriticalDataSource:
    """Critical data source for all species"""

    data: pd.DataFrame
    """Critical data for all species"""

    def __init__(self):
        data: AbstractContextManager[Path] = importlib.resources.as_file(
            DATA_DIRECTORY.joinpath(CRITICAL_DATA_SOURCE)  # type: ignore
        )
        with data as datapath:
            self.data = pd.read_csv(datapath, sep=" ", comment="#")

    @property
    def name_column(self) -> str:
        """Name of the column that refers to the hill formula and an optional suffix"""
        return "name"

    @property
    def critical_temperature_column(self) -> str:
        """Name of the column that refers to the critical temperature in K"""
        return "Tc"

    @property
    def critical_pressure_column(self) -> str:
        """Name of the column that refers to the critical pressure"""
        return "Pc"

    def create_dictionary(self) -> dict[str, CriticalData]:
        """Dictionary of critical data for all species

        Returns:
            Dictionary of critical data for all species
        """
        critical_dict: dict[str, CriticalData] = {}

        for row in self.data.itertuples(index=False):
            name: str = str(getattr(row, self.name_column))
            critical_dict[name] = CriticalData(
                temperature=float(getattr(row, self.critical_temperature_column)),
                pressure=float(getattr(row, self.critical_pressure_column)),
            )

        return critical_dict


# Create dictionaries of instantiated data (JAX-compliant Pytrees) that we can use for lookup.
# It should also be net faster to create these data once and then access (potentially many times).
# These are also set to private to avoid sphinx (autodoc) from printing long strings.
thermodynamic_data_source: ThermodynamicDataSource = ThermodynamicDataSource()
"""Thermodynamic data source

:meta private:
"""
thermodynamic_coefficients_dictionary: dict[str, ThermodynamicCoefficients] = (
    thermodynamic_data_source.create_dictionary()
)
"""Thermodynamic coefficients dictionary

:meta private:
"""
critical_data_source: CriticalDataSource = CriticalDataSource()
"""Critical data source

:meta private:
"""
critical_data_dictionary: dict[str, CriticalData] = critical_data_source.create_dictionary()
"""Critical data dictionary

:meta private:
"""


class IndividualSpeciesData(eqx.Module):
    """Individual species data

    Args:
        formula: Formula
        state: State of aggregation as defined by JANAF
    """

    formula: str
    """Formula"""
    state: str
    """State of aggregation"""
    thermo: ThermodynamicCoefficients
    """Thermodynamic coefficient and methods"""
    composition: ImmutableMap[str, tuple[int, float, float]]
    """Composition"""
    hill_formula: str
    """Hill formula"""
    molar_mass: float = eqx.field(converter=float)
    """Molar mass"""

    def __init__(self, formula: str, state: str):
        self.formula = formula
        self.state = state
        mformula: Formula = Formula(self.formula)
        self.composition = ImmutableMap(mformula.composition().asdict())
        self.hill_formula = mformula.formula
        self.molar_mass = mformula.mass * unit_conversion.g_to_kg
        try:
            self.thermo = thermodynamic_coefficients_dictionary[self.name]
        except KeyError:
            raise KeyError(
                f"{self.name} not available. "
                f"Available species are {thermodynamic_data_source.available_species()}"
            )

    @property
    def elements(self) -> tuple[str, ...]:
        """Elements"""
        return tuple(self.composition.keys())

    @property
    def name(self) -> str:
        """Unique name by combining Hill notation and state of aggregation"""
        return f"{self.hill_formula}_{self.state}"

    def get_gibbs_over_RT(self, temperature: ArrayLike) -> Array:
        """Gets Gibbs energy over RT

        Args:
            temperature: Temperature in K

        Returns:
            Gibbs energy over RT
        """
        return self.thermo.get_gibbs_over_RT(temperature)



'''
Collin Cherubim
June 30, 2025
Plotting script for debug_isofate_coupler.py
'''

import pickle
import time as TIME

import matplotlib.pyplot as plt

from constants import *

# from debug_isofate_coupler_v2 import *
from isofate_coupler import *

# from isofate_coupler_v3 import *
# from isofate_coupler_v3_cannon import *
from isofunks import *
from orbit_params import *

start = TIME.time()

# R_star = 1.0*Rs # [m]
# M_star = 1.0*Ms
# T_star = 6000 # [K]

# # M1 star
# R_star = 0.5*Rs # [m]
# M_star = 0.5*Ms
# T_star = 3600 # [K]

# K5 star
# R_star = 0.7*Rs # [m]
# M_star = 0.7*Ms
# T_star = 4440 # [K]

# # LHS 1140
R_star = 0.22*Rs # [m]
M_star = 0.18*Ms # [kg]
T_star = 3096 # [K]
t_jump = 5.9 - 15.4*(M_star/Ms)
L = 0.0038*Ls

# GJ 3090
# R_star = 0.516*Rs # [m]
# M_star = 0.519*Ms # [kg]
# T_star = 3556 # [K]

# K2-18
# R_star = 0.469*Rs # [m]
# M_star = 0.495*Ms
# T_star = 3500 # [K]

# L = Luminosity(R_star, T_star) # [W]

# Mp = 1.0510228321316297*Me
# P = 1.923431660504492/s2day # 0.2 au for solar type star

# Mp = 2.67899838*Me
# P = 19.76878782/s2day
# f_atm = 1.03428670e-02

# Mp = 1.15662286*Me
# P = 150/s2day
# f_atm = 0.00179197

# O2 planet
# Mp = 10.110718251912926*Me
# P = 1.1579370531867088/s2day
# f_atm = 1.46472048e-01

# O2 world
# Mp = 12.408678941478096*Me
# P = 2.28472886/s2day
# f_atm = 2.17823165e-02

# Mp = 1.1020903707177876*Me
# P = 85.80220254822505/s2day
# f_atm = 0.00134977

# f_atm = 0.04763652467166532
# Mp = 3.2335611753347924*Me
# P = 2.915963713301288/s2day

# K2-18 b
# f_atm = 0.001
# Mp = 8.63*Me
# P = 33/s2day

# # # LHS 1140 b
# f_atm = 0.0129
# Mp = 5.6*Me
# P = 24.74/s2day

# GJ 3090 b
# f_atm = 0.03
# Mp = 3.34*Me
# P = 2.9/s2day
#Rp = 2.13

f_atm = 0.02
Mp = 5*Me
P = 5/s2day

a = SemiMajor(M_star, P) # [m]
Fp = Insolation(L, a)  # [W/m2]
T = EqTemp(Fp, A = 0) # planetary eq temp [K]
F0 = Fp*1e-3 # use for M star 
F_final = 0.170 # 0.170 for GJ 699 MUSCLES; use 0.033 for GJ 1132 MUSCLES
# F0 = Fp*10**(-3.5)*(M_star/Ms) # use for G, K stars
flux_model = 'power law'
stellar_type = 'M1'
t_sat = 5e8 # XUV saturation time [yr]
# t_sat = t_jump*1e9 # XUV saturation time [yr]
d = a # orbital distance [m]
time = 5e9 # total simulation time [yr]
t0 = 1e6 # start time [yr]
t_pms = 0 # pms phase duration [yr]
step_fn = False
mechanism = 'XUV' # if using fixed phi, be sure to change Rp = r_core below and rad_evol = False
RR = True
rad_evol = True
Rp_override = False
n_steps = int(1e5)
n_atmodeller = int(1e4)
thermal = True
M_atm = Mp*f_atm # initial atmospheric mass [kg]
melt_fraction_override = False
save_molecules = True
# mantle_iron_dict = {'type': 'static', 'Fe_mass_fraction': 0.1}
mantle_iron_dict = False

OtoH_enhancement = 1
OtoH_enhanced = OtoH_protosolar*OtoH_enhancement
OtoH_enhanced_mass = OtoH_enhanced*(mu_O/mu_H)

N_He = (HetoH_protosolar_mass/(1 + HetoH_protosolar_mass))*M_atm/mu_He # initial He number [atoms]# N_He = 0
# N_H = (1 - DtoH_solar_mass - OtoH_protosolar_mass - CtoH_protosolar_mass)*M_atm/(1 + HetoH_protosolar_mass)/mu_H  # initial H number [atoms]
N_H = (1 - DtoH_solar_mass - OtoH_enhanced_mass - CtoH_protosolar_mass - NtoH_protosolar_mass - StoH_protosolar_mass)*M_atm/(1 + HetoH_protosolar_mass)/mu_H  # initial H number [atoms]
# N_H = 0
# N_H = M_atm/mu_H
N_D = DtoH_solar_mass*M_atm/(1 + HetoH_protosolar_mass)/mu_D
# N_D = 0
# N_O = OtoH_protosolar_mass*M_atm/(1 + HetoH_protosolar_mass)/mu_O
N_O = OtoH_enhanced_mass*M_atm/(1 + HetoH_protosolar_mass)/mu_O
# N_O = 0
N_C = CtoH_protosolar_mass*M_atm/(1 + HetoH_protosolar_mass)/mu_C
# N_C = 0
mu_avg = (N_H*mu_H + N_He*mu_He + N_D*mu_D + N_O*mu_O + N_C*mu_C)/(N_H+N_He+N_D+N_O+N_C)

print('m_H =', N_H*mu_H)
print('m_He =', N_He*mu_He)
print('m_D =', N_D*mu_D)
print('m_O =', N_O*mu_O)
print('m_C =', N_C*mu_C)

# these print statements serve as a check when running sim.py
print('n_steps =', n_steps)
print('n_atmodeller =', n_atmodeller)
print('f_atm =', f_atm)
print('Mp =', Mp/Me, 'M_earth')
print('F0 =', round(F0, 3), 'W/m2')
print('flux model:', flux_model)
print('stellar type:', stellar_type)
print('P =', P*s2day, 'days')
print('d =', round(a/au2m, 3), 'au')
print('Fp =', round(Fp, 1), 'W/m2')
print('Teq =', round(T, 1), 'K')
print('time =', time/1e9, 'Gyr')
print('mechanism =', mechanism)
print('rad_evol =', rad_evol)
print('mantle_iron_dict', mantle_iron_dict)

# run simulation (from isofate.py)
sol = isocalc(f_atm, Mp, M_star, F0, Fp, T, d, time, mechanism, rad_evol,
N_H = N_H, N_He = N_He, N_D = N_D, N_O = N_O, N_C = N_C, melt_fraction_override = melt_fraction_override,
mu = mu_avg, eps = 0.15, activity = 'medium', flux_model = flux_model, stellar_type = stellar_type, 
Rp_override = Rp_override, t_sat = t_sat, step_fn = step_fn, F_final = F_final, t_pms = t_pms, pms_factor = 1e2,
n_steps = n_steps, t0 = t0, rho_rcb = 1.0, RR = RR,
thermal = thermal, beta = -1.23, n_atmodeller = n_atmodeller, 
save_molecules = save_molecules, mantle_iron_dict = mantle_iron_dict)

# path = '/Users/collin/Documents/Harvard/Research/atm_escape/IsoFATE/monte_carlo/atmodeller/corrected_Psi/transient_D_world_full_isofate'
# outfile = open(path, 'wb')
# pickle.dump(sol, outfile)
# outfile.close()

print('checkpoint 1')

# define solutions for plotting
t_a = sol['time']
# adaptt_a = sol['time_adapted']
rp_a = sol['Rp']
# renv_a = sol['renv']
menv_a = sol['Matm']
vpot_a = sol['Vpot']
fenv_a = sol['fatm']
mloss_a = sol['Mloss']
phi_a = sol['phi']
phic_a = sol['phic']
NH_a = sol['N_H']
NHe_a = sol['N_He']
ND_a = sol['N_D']
NO_a = sol['N_O']
NC_a = sol['N_C']
x1_a = sol['x1']
x2_a = sol['x2']
PhiH_a = sol['Phi_H']
PhiHe_a = sol['Phi_He']
PhiD_a = sol['Phi_D']
PhiO_a = sol['Phi_O']
PhiC_a = sol['Phi_C']
Ts_analytic = sol['T_surf_analytic']
Ts_atmod = sol['T_surf_atmod']
if n_atmodeller != 0:
    n_H2O = sol['atmodeller_final']['H2O_atm']
    n_H2 = sol['atmodeller_final']['H2_atm']
    n_O2 = sol['atmodeller_final']['O2_atm']
    n_CO2 = sol['atmodeller_final']['CO2_atm']
    n_CO = sol['atmodeller_final']['CO_atm']
    n_CH4 = sol['atmodeller_final']['CH4_atm']
    n_total_atm = n_H2O + n_H2 + n_O2 + n_CO2 + n_CO + n_CH4 + NHe_a[-1]/avogadro
    x_CO2 = n_CO2/n_total_atm
    x_CO = n_CO/n_total_atm
    x_CH4 = n_CH4/n_total_atm
    x_H2 = n_H2/n_total_atm
    x_O2 = n_O2/n_total_atm
    x_H2O = n_H2O/n_total_atm

    N_tot = N_H + N_He + N_D + N_O + N_C
    N_tot_molecular = (n_H2O + n_H2 + n_O2 + n_CO + n_CO2 + n_CH4)*avogadro + NHe_a[-1]
    x_H = N_H/N_tot
    x_He = N_He/N_tot
    x_He_molecular = NHe_a[-1]/N_tot_molecular
    x_D = N_D/N_tot
    x_O = N_O/N_tot
    x_C = N_C/N_tot
    print('x_H2O =', x_H2O)
    print('x_H2 =', x_H2)
    print('x_He =', x_He_molecular)
    print('x_O2 =', x_O2)
    print('x_CO2 =', x_CO2)
    print('x_CO =', x_CO)
    print('x_CH4 =', x_CH4)
# calculate mass fraction of He

Y = NHe_a*mu_He/(NH_a*mu_H + NHe_a*mu_He + ND_a*mu_D + NO_a*mu_O + NC_a*mu_C)

print('\n')
print('final f_atm =', fenv_a[-1])
print('final f_atm by species =', (NH_a[-1]*mu_H + NHe_a[-1]*mu_He + ND_a[-1]*mu_D + NO_a[-1]*mu_O + NC_a[-1]*mu_C)/Mp)
print('initial f_atm =', fenv_a[0])
print('initial f_atm by species =', (N_H*mu_H + N_He*mu_He + N_D*mu_D + N_O*mu_O + N_C*mu_C)/Mp)
print('final D/H =', ND_a[-1]/NH_a[-1]/DtoH_solar, '[Solar]')
print('final O/H =', NO_a[-1]/NH_a[-1]/OtoH_protosolar, '[Solar]')
print('final X_He (molar concn) =', NHe_a[-1]/(NH_a[-1] + NHe_a[-1] + ND_a[-1] + NO_a[-1] + NC_a[-1])) # molar concentration
print('final Y_He (mass concn) =', Y[-1]) # mass fraction
print('final planetary radius =', rp_a[-1]/Re, 'R_Earth')
print('\n')

# more planetary properties for analytics
mu = mu_solar
t0 = 1e6/s2yr
r_core = R_core(Mp)
r_env = R_env(Mp, f_atm, Fp, t0)
r_atm = R_atm(T, Mp, r_core, r_env, mu)
Rp = r_core + r_env + r_atm
# Rp = r_core
R_B = R_Bondi(Mp, mu, T) # Bondi radius [m]
R_H = R_Hill(Mp, M_star, d) # Hill radius [m]
# Rp = r_core # use this if rad_evol = False for analytics to match isofate
# Rp = Rp_override
Rp = np.min([Rp, R_B, R_H]) # [m]
# R_avg = (Rp + r_core)/2
# A_avg = 4*np.pi*R_avg**2
# Matm_avg = f_atm*Mp/2
# phi_avg = Matm_avg/A_avg/(time/2)
A = 4*np.pi*Rp**2 # [m2]
# g = G*Mp/Rp**2
M_atm = Mp*f_atm # [kg]


# plotting

# plt.rcParams["figure.figsize"] = (14,3.5)
# # plt.rcParams["figure.figsize"] = (8,8) # for diagnostic
# plt.rcParams.update({'font.size': 14})
# plt.rcParams.update({'axes.titlesize': 11})

##############################################################


plt.rcParams["figure.figsize"] = (14,8)
plt.rcParams.update({'font.size': 14})
plt.rcParams.update({'axes.titlesize': 11})

fig, ((ax1, ax2, ax3), (ax4, ax5, ax6), (ax7, ax8, ax9)) = plt.subplots(3, 3, sharex = False)
# fig, (ax1, ax2, ax3) = plt.subplots(1, 3, sharex = False) # use for just phi, x2, and N1/N2 plots
plt.subplots_adjust(wspace = 0.3)

# phi
g = G*Mp/rp_a**2
H_H = R_gas*T/(M_H*g) # D scale height [m]
H_D = R_gas*T/(M_D*g) # D scale height [m]
ax1.plot(t_a*s2yr, PhiH_a*mu_H, color = 'black', label = 'H flux')
ax1.plot(t_a*s2yr, PhiHe_a*mu_He, color = 'grey', label = 'He flux')
ax1.plot(t_a*s2yr, phic_a, '--', color = 'grey', label = 'He critical')
if ND_a[0] != 0:
    ax1.plot(t_a*s2yr, PhiD_a*mu_D, color = 'orangered', label = 'D flux')
    ax1.plot(t_a*s2yr, b_H_D(T)*x1_a*(mu_D - mu_H)/H_H, '--', color = 'orangered', label = 'D critical') # D/H critical flux
if NO_a[0] != 0:
    ax1.plot(t_a*s2yr, PhiO_a*mu_O, color = 'green', label = 'O flux')
    ax1.plot(t_a*s2yr, b_H_O(T)*x1_a*(mu_O - mu_H)/H_H, '--', color = 'green', label = 'O critical') # O/H critical flux
if NC_a[0] != 0:
    ax1.plot(t_a*s2yr, PhiC_a*mu_C, color = 'gold', label = 'C flux')
    ax1.plot(t_a*s2yr, b_H_C(T)*x1_a*(mu_C - mu_H)/H_H, '--', color = 'gold', label = 'C critical') # C/H critical flux
ax1.plot(t_a*s2yr, phi_a, ':', color = 'mediumslateblue', label = 'total flux')
cross = np.where(np.abs(phi_a - phic_a) < 1e-12)
if len(cross[0]) != 0:
    ax1.axvline(t_a[cross][0]*s2yr, ls = '--', lw = 0.5, color = 'coral')
    print('critical flux at ', t_a[cross][0]*s2yr/1e9, 'Gyr')
ax1.set_ylabel('phi [kg m$^{-2}$ s$^{-1}$]', labelpad = 2)
ax1.legend(fontsize = 8.5, frameon = False, framealpha = 0.7, loc = 3, ncol = 3)
ax1.set_ylim(PhiH_a[0]*mu_H/1e8, PhiH_a[0]*mu_H*100)
ax1.set_yscale('log')

# D/H, O/H
ax2.loglog(t_a*s2yr, x2_a, color = 'maroon', label = 'IsoFATE')
ax2.set_ylabel('N$_{He}$/(N$_H$+N$_{He}$)', labelpad = 2)
# if ND_a[0] != 0:
#     ax2.loglog(t_a*s2yr, ND_a/NH_a, color = 'maroon', label = 'IsoFATE')
#     ax2.set_ylabel('D/H [Solar]', labelpad = 2)
# if NO_a[0] != 0:
#     ax2.loglog(t_a*s2yr, NO_a/NH_a, color = 'maroon', label = 'IsoFATE')
#     ax2.set_ylabel('O/H [Solar]', labelpad = 2)

# ax2.set_title('N$_2$/N$_1$ [xSolar]')
# ax2.set_xlabel('Time [yr]')
# if mechanism == 'fix phi subcritical':
#     x2_analytic = x2sub(x2_0, tau, t_a*s2yr)
#     # x2_analytic = x2sub2(x2_0, N1_0, A, Phi1, t_a*s2yr)
#     ax2.loglog(t_a*s2yr, x2_analytic/dh_a[0,0], '--', color = 'deeppink', zorder = 10, label = 'analytic')
#     ax2.legend(shadow = True)
# elif mechanism == 'fix phi supercritical':
#     x2_analytic = x2super(x2_0, Mp, Rp, T, Phi, tau/s2yr, t_a)
#     ax2.loglog(t_a*s2yr, x2_analytic/dh_a[0,0], '--', color = 'deeppink', zorder = 10, label = 'analytic')
#     ax2.legend(shadow = True)
# else:
    # x2_analytic = analytic_soln(x2_0, N1_0, mu1, mu2, x1_a[0], x2_a[0], Mp, rp_a[0]*Re, fenv_a[0], T, phi_a[0], phic_a[0], t_a)
    # ax2.loglog(t_a*s2yr, x2_analytic[0]/dh_a[0,0], '--', color = 'deeppink', zorder = 10, label = 'analytic')
    # ax2.legend(shadow = True)
# ax2.set_title('atmospheric fractionation')
if len(cross[0]) != 0:
    ax1.axvline(t_a[cross][0]*s2yr, ls = '--', lw = 0.5, color = 'coral')
#ax2.set_xlabel('time [yr]')

# N_x
ax3.plot(t_a*s2yr, NH_a/avogadro, color = 'black', label = 'N$_{H}$')
# ax3.set_xlabel('Time [yr]')
if len(cross[0]) != 0:
    ax3.axvline(t_a[cross][0]*s2yr, ls = '--', lw = 0.5, color = 'coral')
if NHe_a[0] != 0:
    ax3.plot(t_a*s2yr, NHe_a/avogadro, color = 'grey', label = 'N$_{He}$')
if ND_a[0] != 0:
    ax3.plot(t_a*s2yr, ND_a/avogadro, color = 'orangered', label = 'N$_D$')
if NO_a[0] != 0:
    ax3.plot(t_a*s2yr, NO_a/avogadro, color = 'green', label = 'N$_{O}$')
if NC_a[0] != 0:
    ax3.plot(t_a*s2yr, NC_a/avogadro, color = 'gold', label = 'N$_{C}$')
# if mechanism == 'fix phi subcritical' or mechanism == 'fix phi supercritical':
#     ax3.loglog(t_a*s2yr, N1(t_a, A, Phi, N1_0)/avogadro, '--', color = 'skyblue', zorder = 10, label = 'analytic N$_{H}$') # changed H_0 to N1_0
#     ax3.legend(shadow = True)
#else:
    #ax3.loglog(t_a*s2yr, N1(t_a, A, phi_c/mu1, N1_0), '--', color = 'deeppink', zorder = 10, label = 'analytic') # changed H_0 to N1_0
    # ax3.loglog(t_a*s2yr, N1_mod(t_a, 4*np.pi*r_core**2, phi_c/mu1, N1_0), '--', color = 'deeppink', zorder = 10, label = 'analytic') # changed H_0 to N1_0
# ax3.set_title('hydrogen number')
ax3.set_ylabel('atmospheric moles', labelpad = 2)
ax3.legend(frameon = False, fontsize = 9)
ax3.set_ylim(NH_a[0]/avogadro/1e12, NH_a[0]/avogadro*100)
ax3.set_yscale('log')
#ax3.set_xlabel('time [yr]')

# planet radius
ax4.loglog(t_a*s2yr, rp_a/Re, color = 'mediumseagreen')
# ax4.set_title('planetary radius')
ax4.set_ylabel('radius [R$_\oplus$]', labelpad = 2)
#ax4.set_xlabel('time [yr]')

# envelope radius
# ax5.loglog(t_a*s2yr, renv_a[0]*1000, color = 'skyblue')
# # ax5.set_title('envelope radius')
# ax5.set_ylabel('R$_{env}$ [km]', labelpad = 2)
# #ax5.set_xlabel('time [yr]')

ax5.plot(t_a*s2yr, Ts_atmod, color = 'mediumslateblue')
ax5.set_xscale('log')
ax5.set_ylabel('surface temp [K]', labelpad = 2)
ax5.set_ylim(-100, 6100)
ax5.annotate(f'Rp = {round(rp_a[-1]/Re, 2)} Re, Mp = {round(Mp/Me, 2)} Me, Teq = {round(T, 0)} K', (1e6, 5500), fontsize = 10)
ax5.annotate('final fatm:'+str(round(fenv_a[-1], 6)), (1e6, 5000), fontsize = 10)
ax5.annotate('final D/H:'+str(round(ND_a[-1]/NH_a[-1]/DtoH_solar, 2))+' [Solar]', (1e6, 4500), fontsize = 10)
if n_atmodeller != 0:
    ax5.annotate('final He concn:'+str(round(x_He_molecular, 4))+'='+str(round(Y[-1], 4))+' kg/kg', (1e6, 4000), fontsize = 10)
    ax5.annotate('final O2 concn:'+str(round(x_O2, 4)), (1e6, 3500), fontsize = 10)
    ax5.annotate('final H2O concn:'+str(round(x_H2O, 4)), (1e6, 3000), fontsize = 10)
    ax5.annotate('final CO2 concn:'+str(round(x_CO2, 4)), (1e6, 2500), fontsize = 10)
    ax5.annotate('final CO concn:'+str(round(x_CO, 4)), (1e6, 2000), fontsize = 10)
    ax5.annotate('final CH4 concn:'+str(round(x_CH4, 4)), (1e6, 1500), fontsize = 10)
    ax5.annotate('final H2 concn:'+str(round(x_H2, 4)), (1e6, 1000), fontsize = 10)

# envelope mass
ax6.loglog(t_a*s2yr, menv_a/Me, color = 'midnightblue')
# ax6.set_title('envelope mass')
ax6.set_ylabel('M$_{env}$ [M$_\oplus$]', labelpad = 2)
#ax6.set_xlabel('time [yr]')

# grav potential
ax7.loglog(t_a*s2yr, vpot_a, color = 'orange')
# ax7.set_title('gravitational potential')
ax7.set_ylabel('V$_{pot}$ [J/kg]', labelpad = 2)
ax7.set_xlabel('time [yr]')

# mass loss
ax8.loglog(t_a*s2yr, mloss_a, color = 'crimson')
# ax8.set_title('mass loss per time step')
ax8.set_ylabel('$\Delta$ mass [kg]', labelpad = 2)
ax8.set_xlabel('time [yr]')

# f_env or system parameters
ax9.set_xlabel('time [yr]')

### model parameters

# if adaptive_steps == False:
# ax9.annotate('f_atm ='+str(f_atm)+',', (1e8, 0.9))
# ax9.annotate('comp:'+str(species), (3e9, 0.9))
# ax9.annotate('Mp ='+str(round(Mp/Me, 2))+' M$_\oplus$', (1e8, 0.78))
# ax9.annotate('Rp(t=0) ='+str(round(Rp/Re, 2))+' R$_\oplus$', (1e8, 0.66))
# ax9.annotate('F$_{XUV}$(t=0) ='+str(round(F0, 2))+' W/m2', (1e8, 0.54))
# ax9.annotate('P ='+str(round(P*s2day, 2))+' days', (1e8, 0.42))
# ax9.annotate('Fp ='+str(round(Fp, 2))+' W/m2', (1e8, 0.30))
# ax9.annotate('Teq ='+str(round(T, 2))+' K', (1e8, 0.18))
# ax9.annotate('mechanism ='+str(mechanism), (1e8, 0.06))
# ax9.tick_params(left = 'False')
# ax9.set_yticks([])

### atmospheric mass fraction

ax9.set_ylabel('f_env [%]', labelpad = 2)
ax9.plot(t_a*s2yr, fenv_a*100, color = 'gold')
ax9.set_xscale('log')

plt.tight_layout()




print('done (', round((TIME.time() - start)/60, 2), 'mins )')

# plt.savefig('/Users/collin/Documents/Harvard/Research/atm_escape/IsoFATE/case_studies/LHS1140b_f0017_Fi1_tjump_Ffinal170_XUV+RR_atmod1e2_ntime1e5_t10e9_tpms3e8_v2.png', dpi = 300, bbox_inches = 'tight')

plt.show();



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for the EOS models from :cite:t:`Lide2005`"""

from atmodeller.eos import RealGas

MODEL_SUFFIX: str = "vdw_lide05"
"""Suffix of the :cite:t:`Lide2005` models"""


def test_H2_volume_100bar(check_values) -> None:
    """Tests H2 volume at 300 K and 100 bar"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    check_values.volume(300, 100, model, 2.85325715e-4)


def test_H2_volume_1kbar(check_values) -> None:
    """Tests H2 volume at 1000 K and 1000 bar"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    check_values.volume(1000, 1000, model, 1.12342095e-4)


def test_volume_with_broadcasting(check_values) -> None:
    """Tests volume with broadcasting"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    check_values.check_broadcasting("volume", model)


def test_fugacity_with_broadcasting(check_values) -> None:
    """Tests volume with broadcasting"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    check_values.check_broadcasting("fugacity", model)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests redox buffers"""

import logging

from jaxtyping import ArrayLike
from pytest import approx

from atmodeller import debug_logger
from atmodeller.thermodata._redox_buffers import (
    IronWustiteBufferHirschmann,
    IronWustiteBufferHirschmann08,
    IronWustiteBufferHirschmann21,
    RedoxBuffer,
)
from atmodeller.utilities import unit_conversion

logger: logging.Logger = debug_logger()
logger.setLevel(logging.WARNING)

LOG10_SHIFT: ArrayLike = 0
"""Log10 shift"""
RTOL: float = 1.0e-8
"""Relative tolerance"""
ATOL: float = 1.0e-8
"""Absolute tolerance"""

TEST_LOW_TEMPERATURE: ArrayLike = 1000
TEST_LOW_PRESSURE: ArrayLike = 1  # bar
TEST_MEDIUM_TEMPERATURE: ArrayLike = 2000
TEST_MEDIUM_PRESSURE: ArrayLike = 5 * unit_conversion.GPa_to_bar
TEST_HIGH_TEMPERATURE: ArrayLike = 3000
TEST_HIGH_PRESSURE: ArrayLike = 50 * unit_conversion.GPa_to_bar

# The switch temperature is contained in the temperature_min attribute of the calibration class
# within the high temperature IW buffer. The value is 1000 K, which these constants bridge.
TEST_BELOW_SWITCH_TEMPERATURE: ArrayLike = 900
TEST_ABOVE_SWITCH_TEMPERATURE: ArrayLike = 1100


def test_IW_Hirschmann08() -> None:
    """Tests the low temperature iron-wustite buffer :cite:p:`OP93,HGD08`"""
    buffer: RedoxBuffer = IronWustiteBufferHirschmann08(LOG10_SHIFT, None)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(
        TEST_BELOW_SWITCH_TEMPERATURE, TEST_LOW_PRESSURE
    )
    logger.info("log10_fugacity (Very low T, low P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-23.93938230619323, RTOL, ATOL)


def test_IW_Hirschmann21() -> None:
    """Tests the high temperature iron-wustite buffer :cite:p:`H21`

    The values below agree with the calculator provided in Table S7 of the Excel spreadsheet
    provided in the supplementary materials of the online publication.
    """
    buffer: RedoxBuffer = IronWustiteBufferHirschmann21(LOG10_SHIFT, None)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_LOW_TEMPERATURE, TEST_LOW_PRESSURE)
    logger.info("log10_fugacity (Low T, low P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-20.816597461900887, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_MEDIUM_TEMPERATURE, TEST_LOW_PRESSURE)
    logger.info("log10_fugacity (Med T, low P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-7.053620769567622, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_HIGH_TEMPERATURE, TEST_LOW_PRESSURE)
    logger.info("log10_fugacity (High T, low P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-2.562340960147525, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_LOW_TEMPERATURE, TEST_MEDIUM_PRESSURE)
    logger.info("log10_fugacity (Low T, med P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-17.97502528402518, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(
        TEST_MEDIUM_TEMPERATURE, TEST_MEDIUM_PRESSURE
    )
    logger.info("log10_fugacity (Med T, med P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-5.619533227830953, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_HIGH_TEMPERATURE, TEST_MEDIUM_PRESSURE)
    logger.info("log10_fugacity (High T, med P) = %s", log10_fugacity)

    assert log10_fugacity == approx(-1.5951793921085295, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_LOW_TEMPERATURE, TEST_HIGH_PRESSURE)
    logger.info("log10_fugacity (Low T, high P) = %s", log10_fugacity)

    assert log10_fugacity == approx(4.232976104931987, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_MEDIUM_TEMPERATURE, TEST_HIGH_PRESSURE)
    logger.info("log10_fugacity (Med T, high P) = %s", log10_fugacity)

    assert log10_fugacity == approx(5.534585436736885, RTOL, ATOL)

    log10_fugacity: ArrayLike = buffer.log10_fugacity(TEST_HIGH_TEMPERATURE, TEST_HIGH_PRESSURE)
    logger.info("log10_fugacity (High T, high P) = %s", log10_fugacity)

    assert log10_fugacity == approx(5.887511178821558, RTOL, ATOL)


def test_IW_Hirschmann() -> None:
    """Tests the switch between the high and low temperature iron-wustite buffer."""

    low_temperature_buffer: RedoxBuffer = IronWustiteBufferHirschmann08(LOG10_SHIFT, None)
    high_temperature_buffer: RedoxBuffer = IronWustiteBufferHirschmann21(LOG10_SHIFT, None)
    composite_buffer: RedoxBuffer = IronWustiteBufferHirschmann(LOG10_SHIFT, None)

    # Evaluate the buffers either side of the switch temperature.
    log10_fugacity_low_temperature: ArrayLike = low_temperature_buffer.log10_fugacity(
        TEST_BELOW_SWITCH_TEMPERATURE, TEST_LOW_PRESSURE
    )
    log10_fugacity_composite: ArrayLike = composite_buffer.log10_fugacity(
        TEST_BELOW_SWITCH_TEMPERATURE, TEST_LOW_PRESSURE
    )

    assert log10_fugacity_composite == approx(log10_fugacity_low_temperature, RTOL, ATOL)

    log10_fugacity_high_temperature: ArrayLike = high_temperature_buffer.log10_fugacity(
        TEST_ABOVE_SWITCH_TEMPERATURE, TEST_LOW_PRESSURE
    )
    log10_fugacity_composite: ArrayLike = composite_buffer.log10_fugacity(
        TEST_ABOVE_SWITCH_TEMPERATURE, TEST_LOW_PRESSURE
    )

    assert log10_fugacity_composite == approx(log10_fugacity_high_temperature, RTOL, ATOL)


def test_IW_Hirschmann08_fixed_P() -> None:
    """Tests the low temperature iron-wustite buffer with constant pressure

    When the user sets an evaluation pressure the total pressure should not enter the calculation
    of the fugacity.
    """
    # Define an evaluation pressure at 1 bar
    buffer: RedoxBuffer = IronWustiteBufferHirschmann08(LOG10_SHIFT, 1)

    log10_fugacity_low_pressure: ArrayLike = buffer.log10_fugacity(
        TEST_BELOW_SWITCH_TEMPERATURE, TEST_LOW_PRESSURE
    )
    logger.info("log10_fugacity_low_pressure = %s", log10_fugacity_low_pressure)

    log10_fugacity_high_pressure: ArrayLike = buffer.log10_fugacity(
        TEST_BELOW_SWITCH_TEMPERATURE, TEST_HIGH_PRESSURE
    )
    logger.info("log10_fugacity_high_pressure = %s", log10_fugacity_high_pressure)

    # The evaluation should be at 1 bar pressure and the input pressure ignored
    assert log10_fugacity_low_pressure == approx(-23.93938230619323, RTOL, ATOL)
    assert log10_fugacity_high_pressure == approx(-23.93938230619323, RTOL, ATOL)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Classes"""

import logging
import pprint
from collections.abc import Callable, Mapping
from typing import Any, Optional, cast

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as random
import numpy as np
from jaxtyping import Array, ArrayLike, Bool, Float, Integer, PRNGKeyArray

from atmodeller import INITIAL_LOG_NUMBER_DENSITY, INITIAL_LOG_STABILITY, TAU, TAU_MAX, TAU_NUM
from atmodeller._mytypes import NpFloat, NpInt
from atmodeller.containers import (
    FixedParameters,
    FugacityConstraints,
    MassConstraints,
    Planet,
    SolverParameters,
    SpeciesCollection,
    TracedParameters,
)
from atmodeller.engine import make_solve_tau_step, repeat_solver, solve
from atmodeller.interfaces import FugacityConstraintProtocol
from atmodeller.output import Output
from atmodeller.utilities import get_batch_size, partial_rref, vmap_axes_spec

logger: logging.Logger = logging.getLogger(__name__)


class InteriorAtmosphere:
    """Interior atmosphere

    This is the main class that the user interacts with to build interior-atmosphere systems,
    solve them, and retrieve the results.

    Args:
        species: Collection of species
        tau: Tau factor for species stability. Defaults to TAU.
    """

    _solver: Optional[Callable] = None
    _output: Optional[Output] = None

    def __init__(self, species: SpeciesCollection, tau: float = TAU):
        self.species: SpeciesCollection = species
        self.tau: float = tau
        logger.info("species = %s", str(self.species))
        logger.info("reactions = %s", pprint.pformat(self.get_reaction_dictionary()))

    @property
    def output(self) -> Output:
        if self._output is None:
            raise AttributeError("Output has not been set.")

        return self._output

    def solve(
        self,
        *,
        planet: Optional[Planet] = None,
        initial_log_number_density: Optional[ArrayLike] = None,
        initial_log_stability: Optional[ArrayLike] = None,
        fugacity_constraints: Optional[Mapping[str, FugacityConstraintProtocol]] = None,
        mass_constraints: Optional[Mapping[str, ArrayLike]] = None,
        solver_parameters: Optional[SolverParameters] = None,
    ) -> None:
        """Solves the system and initialises an Output instance for processing the result

        Args:
            planet: Planet. Defaults to None.
            initial_log_number_density: Initial log number density. Defaults to None.
            initial_log_stability: Initial log stability. Defaults to None.
            fugacity_constraints: Fugacity constraints. Defaults to None.
            mass_constraints: Mass constraints. Defaults to None.
            solver_parameters: Solver parameters. Defaults to None.
        """
        planet_: Planet = Planet() if planet is None else planet

        batch_size: int = get_batch_size((planet, fugacity_constraints, mass_constraints))

        fugacity_constraints_: FugacityConstraints = FugacityConstraints.create(
            self.species, fugacity_constraints
        )
        mass_constraints_: MassConstraints = MassConstraints.create(
            self.species, mass_constraints, batch_size
        )

        # Always broadcast tau because the repeat_solver is triggered if some cases fail
        broadcasted_tau: Float[Array, " batch_dim"] = jnp.full((batch_size,), TAU)
        # jax.debug.print("broadcasted_tau = {out}", out=broadcasted_tau)

        traced_parameters_: TracedParameters = TracedParameters(
            planet_, fugacity_constraints_, mass_constraints_
        )
        fixed_parameters_: FixedParameters = self.get_fixed_parameters()
        solver_parameters_: SolverParameters = (
            SolverParameters() if solver_parameters is None else solver_parameters
        )
        options: dict[str, Any] = {
            "lower": self.species.get_lower_bound(),
            "upper": self.species.get_upper_bound(),
            "jac": solver_parameters_.jac,
        }

        # NOTE: Determine active entries in the residual. This order must correspond to the order
        # of entries in the residual.
        active: Bool[Array, " res_dim"] = jnp.concatenate(
            (
                fugacity_constraints_.active(),
                fixed_parameters_.active_reactions(),
                mass_constraints_.active(),
                fixed_parameters_.active_stability(),
            )
        )
        # jax.debug.print("active = {out}", out=active)
        active_indices: Integer[Array, "..."] = jnp.where(active)[0]
        # jax.debug.print("active_indices = {out}", out=active_indices)

        base_solution_array: Array = broadcast_initial_solution(
            initial_log_number_density,
            initial_log_stability,
            self.species.number,
            batch_size,
        )
        # jax.debug.print("base_solution_array = {out}", out=base_solution_array)

        # Pre-bind fixed configurations
        solver_fn: Callable = eqx.Partial(
            solve,
            fixed_parameters=fixed_parameters_,
            options=options,
        )
        in_axes: TracedParameters = vmap_axes_spec(traced_parameters_)

        # Compile the solver, and this is re-used unless recompilation is triggered
        # Initial solution and tau must be broadcast since they are always batched
        self._solver = eqx.filter_jit(
            eqx.filter_vmap(solver_fn, in_axes=(0, None, 0, in_axes, None))
        )

        # First solution attempt. If the initial guess is close enough we might just find solutions
        # for all cases.
        logger.info(f"Attempting to solve {batch_size} model(s)")
        solution, solver_status, solver_steps = self._solver(
            base_solution_array,
            active_indices,
            broadcasted_tau,
            traced_parameters_,
            solver_parameters_,
        )
        # jax.debug.print("solver_status = {out}", out=solver_status)
        # jax.debug.print("solver_steps = {out}", out=solver_steps)
        solver_attempts: Integer[Array, " batch_dim"] = solver_status.astype(int)
        # jax.debug.print("solver_attempts = {out}", out=solver_attempts)

        if jnp.any(~solver_status):
            num_failed: int = jnp.sum(~solver_status).item()
            logger.warning("%d model(s) failed to converge on the first attempt", num_failed)
            logger.warning(
                "But don't panic! This can happen when starting from a poor initial guess."
            )
            logger.warning(
                "Launching multistart (maximum %d attempts)", solver_parameters_.multistart
            )
            logger.warning(
                "Attempting to solve the %d models(s) that initially failed", num_failed
            )

            # Restore the base solution for cases that failed since this will be perturbed
            solution: Float[Array, "batch_dim sol_dim"] = cast(
                Array, jnp.where(solver_status[:, None], solution, base_solution_array)
            )
            # jax.debug.print("solution = {out}", out=solution)

            # Use repeat solver to ensure all cases solve
            key: PRNGKeyArray = jax.random.PRNGKey(0)
            key, subkey = random.split(key)

            # Prototyping switching the solver for
            # solver_parameters_ = eqx.tree_at(
            #     lambda sp: sp.solver,
            #     solver_parameters_,  # your original instance
            #     optx.LevenbergMarquardt,  # or whatever solver you want to use
            # )
            # print(new_solver_params)

            if jnp.any(fixed_parameters_.active_stability()):
                logger.info(
                    "Multistart with species' stability (TAU_MAX= %.1e, TAU= %.1e, TAU_NUM= %d)",
                    TAU_MAX,
                    TAU,
                    TAU_NUM,
                )
                varying_tau_row: Float[Array, " tau_dim"] = jnp.logspace(
                    jnp.log10(TAU_MAX), jnp.log10(TAU), num=TAU_NUM
                )
                constant_tau_row: Float[Array, " tau_dim"] = jnp.full((TAU_NUM,), TAU)
                tau_templates: Float[Array, "tau_dim 2"] = jnp.stack(
                    [varying_tau_row, constant_tau_row], axis=1
                )
                tau_array: Float[Array, "tau_dim batch_dim"] = tau_templates[
                    :, solver_status.astype(int)
                ]
                # jax.debug.print("tau_array = {out}", out=tau_array)

                initial_carry: tuple[Array, Array] = (subkey, solution)
                solve_tau_step: Callable = make_solve_tau_step(
                    self._solver, active_indices, traced_parameters_, solver_parameters_
                )
                _, results = jax.lax.scan(solve_tau_step, initial_carry, tau_array)
                solution, solver_status_, solver_steps_, solver_attempts = results

                # Debugging output. Requires the complete arrays as given above.
                failed_indices: Integer[Array, "..."] = jnp.where(~solver_status)[0]
                for ii in failed_indices.tolist():
                    logger.debug(f"--- Solve summary for failed index {ii} ---")
                    for tau_i in range(TAU_NUM):
                        status_i: bool = bool(solver_status_[tau_i, ii])
                        steps_i: int = int(solver_steps_[tau_i, ii])
                        attempts_i: int = int(solver_attempts[tau_i, ii])
                        logger.debug(
                            "Tau step %1d: status= %-5s  steps= %3d  attempts= %2d",
                            tau_i,
                            str(status_i),
                            steps_i,
                            attempts_i,
                        )

                # Aggregate output
                solution = solution[-1]  # Only need solution for final TAU
                solver_status_ = solver_status_[-1]  # Only need status for final TAU
                solver_steps_ = jnp.sum(solver_steps_, axis=0)  # Sum steps for all tau
                solver_attempts = jnp.max(solver_attempts, axis=0)  # Max for all tau

                # jax.debug.print("solution = {out}", out=solution)
                # jax.debug.print("solver_status_ = {out}", out=solver_status_)
                # jax.debug.print("solver_steps_ = {out}", out=solver_steps_)
                # jax.debug.print("solver_attempts = {out}", out=solver_attempts)

                # Maximum attempts across all tau and all models
                max_attempts: int = jnp.max(solver_attempts).item()

            else:
                solution, solver_status_, solver_steps_, solver_attempts = repeat_solver(
                    self._solver,
                    active_indices,
                    broadcasted_tau,
                    solution,
                    traced_parameters_,
                    solver_parameters_,
                    subkey,
                )
                max_attempts = jnp.max(solver_attempts).item()
                # Since tau is unaltered, the first multistart just repeats the first calculation,
                # which we already know has some failed cases. So we minus one for the reporting.
                max_attempts -= 1

            logger.info("Multistart complete with %s total attempt(s)", max_attempts)

            # Restore statistics of cases that solved first time
            solver_steps: Integer[Array, " batch_dim"] = jnp.where(
                solver_status, solver_steps, solver_steps_
            )
            solver_status: Bool[Array, " batch_dim"] = solver_status_  # Final status

            # Count unique values and their frequencies
            unique_vals, counts = jnp.unique(solver_attempts, return_counts=True)
            for val, count in zip(unique_vals.tolist(), counts.tolist()):
                logger.info(
                    "Multistart, max attempts: %d, model count: %d (%0.2f%%)",
                    val,
                    count,
                    count * 100 / batch_size,
                )

        num_successful_models: int = jnp.count_nonzero(solver_status).item()
        num_failed_models: int = jnp.count_nonzero(~solver_status).item()

        logger.info(
            "Solve complete: %d (%0.2f%%) successful model(s)",
            num_successful_models,
            num_successful_models * 100 / batch_size,
        )

        if num_failed_models > 0:
            logger.warning(
                "%d (%0.2f%%) model(s) still failed",
                num_failed_models,
                num_failed_models * 100 / batch_size,
            )

        logger.info("Solver steps (max) = %s", jnp.max(solver_steps).item())

        self._output = Output(
            self.species,
            solution,
            active_indices,
            solver_status,
            solver_steps,
            solver_attempts,
            fixed_parameters_,
            traced_parameters_,
            solver_parameters_,
        )

    def get_fixed_parameters(self) -> FixedParameters:
        """Gets fixed parameters.

        Returns:
            Fixed parameters
        """
        formula_matrix: NpInt = self.get_formula_matrix()
        reaction_matrix: NpFloat = self.get_reaction_matrix()
        gas_species_mask: Array = self.species.get_gas_species_mask()
        molar_masses: Array = self.species.get_molar_masses()
        diatomic_oxygen_index: int = self.species.get_diatomic_oxygen_index()

        fixed_parameters: FixedParameters = FixedParameters(
            species=self.species,
            formula_matrix=jnp.asarray(formula_matrix),
            reaction_matrix=jnp.asarray(reaction_matrix),
            gas_species_mask=gas_species_mask,
            diatomic_oxygen_index=diatomic_oxygen_index,
            molar_masses=molar_masses,
        )

        return fixed_parameters

    def get_formula_matrix(self) -> NpInt:
        """Gets the formula matrix.

        Elements are given in rows and species in columns following the convention in
        :cite:t:`LKS17`.

        Returns:
            Formula matrix
        """
        unique_elements: tuple[str, ...] = self.species.get_unique_elements_in_species()
        formula_matrix: NpInt = np.zeros(
            (len(unique_elements), self.species.number), dtype=np.int_
        )

        for element_index, element in enumerate(unique_elements):
            for species_index, species_ in enumerate(self.species):
                count: int = 0
                try:
                    count = species_.data.composition[element][0]
                except KeyError:
                    count = 0
                formula_matrix[element_index, species_index] = count

        # logger.debug("formula_matrix = %s", formula_matrix)

        return formula_matrix

    def get_reaction_matrix(self) -> NpFloat:
        """Gets the reaction matrix.

        Returns:
            A matrix of linearly independent reactions or an empty array if no reactions
        """
        if self.species.number == 1:
            logger.debug("Only one species therefore no reactions")
            return np.array([], dtype=np.float64)

        transpose_formula_matrix: NpInt = self.get_formula_matrix().T
        reaction_matrix: NpFloat = partial_rref(transpose_formula_matrix)

        return reaction_matrix

    def get_reaction_dictionary(self) -> dict[int, str]:
        """Gets reactions as a dictionary.

        Returns:
            Reactions as a dictionary
        """
        reaction_matrix: NpFloat = self.get_reaction_matrix()
        reactions: dict[int, str] = {}
        if reaction_matrix.size != 0:
            for reaction_index in range(reaction_matrix.shape[0]):
                reactants: str = ""
                products: str = ""
                for species_index, species_ in enumerate(self.species):
                    coeff: float = reaction_matrix[reaction_index, species_index].item()
                    if coeff != 0:
                        if coeff < 0:
                            reactants += f"{abs(coeff)} {species_.data.name} + "
                        else:
                            products += f"{coeff} {species_.data.name} + "

                reactants = reactants.rstrip(" + ")
                products = products.rstrip(" + ")
                reaction: str = f"{reactants} = {products}"
                reactions[reaction_index] = reaction

        return reactions


def _broadcast_component(
    component: Optional[ArrayLike], default_value: float, dim: int, batch_size: int, name: str
) -> NpFloat:
    """Broadcasts a scalar, 1D, or 2D input array to shape (batch_size, dim).

    This function standardizes inputs that may be:
        - None (in which case a default value is used),
        - a scalar (promoted to a 1D array of length `dim`),
        - a 1D array of shape (`dim`,) (broadcast across the batch),
        - or a 2D array of shape (`batch_size`, `dim`) (used as-is).

    Args:
        component: The input data (or None), representing either a scalar, 1D array, or 2D array
        default_value: The default scalar value to use if `component` is None
        dim: The number of features or dimensions per batch item
        batch_size: The number of batch items
        name: Name of the component (used for error messages)

    Returns:
        A numpy array of shape (batch_size, dim), with values broadcast as needed

    Raises:
        ValueError: If the input array has an unexpected shape or inconsistent dimensions
    """
    if component is None:
        base: NpFloat = np.full((dim,), default_value, dtype=np.float64)
    else:
        component = np.asarray(component, dtype=jnp.float64)
        if component.ndim == 0:
            base = np.full((dim,), component.item(), dtype=np.float64)
        elif component.ndim == 1:
            if component.shape[0] != dim:
                raise ValueError(f"{name} should have shape ({dim},), got {component.shape}")
            base = component
        elif component.ndim == 2:
            if component.shape[0] != batch_size or component.shape[1] != dim:
                raise ValueError(
                    f"{name} should have shape ({batch_size}, {dim}), got {component.shape}"
                )
            # Replace NaNs with default_value
            component = np.where(np.isnan(component), default_value, component)
            return component
        else:
            raise ValueError(
                f"{name} must be a scalar, 1D, or 2D array, got shape {component.shape}"
            )

    # Promote 1D base to (batch_size, dim)
    return np.broadcast_to(base[None, :], (batch_size, dim))


def broadcast_initial_solution(
    initial_log_number_density: Optional[ArrayLike],
    initial_log_stability: Optional[ArrayLike],
    number_of_species: int,
    batch_size: int,
) -> Array:
    """Creates and broadcasts the initial solution to shape (batch_size, D)

    D = number_of_species + number_of_stability, i.e. the total number of solution quantities

    Args:
        initial_log_number_density: Initial log number density. Defaults to None.
        initial_log_stability: Initial log stability. Defaults to None.
        number_of_species: Number of species
        batch_size: Batch size

    Returns:
        Initial solution with shape (batch_size, D)
    """
    number_density: NpFloat = _broadcast_component(
        initial_log_number_density,
        INITIAL_LOG_NUMBER_DENSITY,
        number_of_species,
        batch_size,
        name="initial_log_number_density",
    )
    stability: NpFloat = _broadcast_component(
        initial_log_stability,
        INITIAL_LOG_STABILITY,
        number_of_species,
        batch_size,
        name="initial_log_stability",
    )

    return jnp.concatenate((number_density, stability), axis=-1)



'''
Collin Cherubim
October 2, 2022
Frequently used physical constants
'''

inv_cm2m = 100 # convert inverse cm to inverse m
avogadro = 6.022e23 # Avogadro's number [particles/mole]
s2day = 1/(3600*24) # convert s to day
s2yr = 1/(3600*24*365) # convert s to year
au2m = 1.496e11 # convert au to m
cgs2si_flux = 1/1000 # convert flux from erg/cm2/s to W/m2
erg2joule = 1e-7 # convert ergs to Joules
gcm2kgm = 1000 # convert g/cm3 to kg/m3 (SI)
J2E_mass = 1.898e27/5.972e24 # convert Jupiter mass to Earth mass [kg]
J2E_rad = 7.1492e7/6.3781e6 # convert Jupiter radius to Earth radius [m]
R_gas = 8.314462 # gas constant [J/mol/K]
kb = 1.38e-23 # Boltzmann constant [m2 kg/s2 K]
sbc = 5.67e-8 # Stefan Boltzmann constant W/m2/K4
g_e = 9.8 # grav field strength [m/s/s]
G = 6.6743e-11 # [m3/kg/s2]
h = 6.62607015e-34 # [J s]
c = 2.99792458e8 # [m/s]
Re = 6.378e6 # Earth radius [m]
Me = 5.9722e24 # Earth mass [kg]
Fe = 1366 # Earth bolometric flux [W/m2]
Ms = 1.98847e30 # Solar mass [kg]
Rs = 6.957e8 # Solar radius [m]
Ls = 3.828e26 # Solar luminosity [W]
Mjup = 1.898e27 # Jupiter mass [kg]
Rjup = 7.1492e7 # Jupiter radius [m]
M_N2 = 0.028014 # molar mass N2 [kg/mol]
M_H2 = 0.002016 # molar mass H2 [kg/mol]
M_HD = 0.003024 # molar mass HD [kg/mol]
M_D = 0.002014 # molar mass of D [kg/mol]
M_H  = 0.001008 # molar mass of H [kg/mol]
M_He = 0.0040026 # molar mass of He [kg/mol]
M_O = 0.015999 # molar mass of O [kg/mol]
M_C = 0.012011 # molar mass of C [kg/mol]
M_N = 0.014007 # molar mass of N [kg/mol]
M_S = 0.032065 # molar mass of S [kg/mol]
M_Fe = 0.055845 # molar mass of Fe [kg/mol]
M_H2O = 0.018015 # molar mass of H2O [kg/mol]
mu_H2 = M_H2/avogadro # molecular mass of H2 [kg/molecule]
mu_HD = M_HD/avogadro # molecular mass of HD [kg/molecule]
mu_H = M_H/avogadro # atomic mass of H [kg/atom]
mu_D = M_D/avogadro # atomic mass of D [kg/atom]
mu_He = M_He/avogadro # atomic mass of He [kg/atom]
mu_O = M_O/avogadro # atomic mass of O [kg/atom]
mu_C = M_C/avogadro # atomic mass of C [kg/atom]
mu_N = M_N/avogadro # atomic mass of N [kg/atom]
mu_S = M_S/avogadro # atomic mass of S [kg/atom]
mu_Fe = M_Fe/avogadro # atomic mass of Fe [kg/atom]
vsmow = 1.5574e-4 # Vienna Standard Mean Ocean Water (D/H for Earth's oceans)
DtoH_solar = 0.0000194 # D/H solar mole ratio from Lodders 2003
DtoH_solar_mass = DtoH_solar*(mu_D/mu_H) # D/H solar mass ratio from Lodders 2003
HetoH_solar = 0.07991 # He/H solar mole ratio from Lodders 2003 (0.2377/0.7491*mu_H/mu_He)
HetoH_protosolar = 0.09709 # He/H proto-solar mole ratio Lodders 2003 (0.2741/0.711*mu_H/mu_He)
HetoH_protosolar_mass = 0.38551 # He/H proto-solar mass ratio Lodders 2003 (0.2741/0.711)
OtoH_protosolar = 0.00058 # O/H proto-solar mole ratio from Lodders 2003 Table 2 (1.413e7/2.431e10)
OtoH_protosolar_mass = OtoH_protosolar*(mu_O/mu_H) # O/H proto-solar mass ratio from Lodders 2003
CtoH_protosolar = 0.00029 # C/H proto-solar mole ratio from Lodders 2003 Table 2
CtoH_protosolar_mass = CtoH_protosolar*(mu_C/mu_H) # O/H proto-solar mass ratio from Lodders 2003
NtoH_protosolar = 0.000080 # N/H proto-solar mole ratio from Lodders 2003 Table 2 (1.950e6/2.431e10)
NtoH_protosolar_mass = NtoH_protosolar*(mu_N/mu_H) # O/H proto-solar mass ratio
StoH_protosolar = 0.000018 # S/H proto-solar mole ratio from Lodders 2003 Table 2 (4.449e5/2.431e10)
StoH_protosolar_mass = StoH_protosolar*(mu_S/mu_H) # O/H proto-solar mass ratio 
mu_HHe = 0.00122/avogadro # H/He with solar abundances
mu_H2He = 0.00227/avogadro # H2/He with solar abundances
mu_solar = 0.00235/avogadro # average particle mass for solar metallicity
n_OperTO = 7.83e22 # mols O per TO
Venus_TO = 1.32e-5 # number of terrestrial oceans in Venus' atmosphere (needs checking)
def b_H_D(T):
    return 7.183e19*T**0.728 # [molecules/m/s] from Genda & Ikoma 2008 for D in H (not measured directly)
def b_H_He(T):
    return 1.04e20*T**0.732 # [molecules/m/s] from Mason & Marrero 1970 for H in He
def b_He_D(T):
    return 5.087e19*T**0.728 # [molecules/m/s] approximated from b_H_D using Genda/Ikoma 2008 prescription (Appendix C)
def b_H_O(T):
    return 4.8e19*T**0.75 # [molecules/m/s] from Wordsworth et al 2018
def b_He_O(T):
    return 2.61e19*T**0.75 # [molecules/m/s] approximated from b_H_O using Genda/Ikoma 2008 prescription (Appendix C)
def b_H_C(T):
    return 4.85e19*T**0.75 # [molecules/m/s] approximated from b_H_O using Genda/Ikoma 2008 prescription (Appendix C)
def b_He_C(T):
    return 2.64e19*T**0.75 # [molecules/m/s] approximated from b_He_O using Genda/Ikoma 2008 prescription (Appendix C)
def b_H_N(T):
    return 4.85e19*T**0.75 # [molecules/m/s] approximated from b_H_O using Genda/Ikoma 2008 prescription (Appendix C)
def b_He_N(T):  
    return 2.65e19*T**0.75 # [molecules/m/s] approximated from b_He_O using Genda/Ikoma 2008 prescription (Appendix C)
def b_H_S(T):
    return 4.73e19*T**0.75 # [molecules/m/s] approximated from b_H_O using Genda/Ikoma 2008 prescription (Appendix C)
def b_He_S(T):
    return 2.48e19*T**0.75 # [molecules/m/s] approximated from b_He_O using Genda/Ikoma 2008 prescription (Appendix C)


#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for the EOS models from :cite:t:`SF87,SF87a,SF88,SS92`"""

from atmodeller.eos import RealGas
from atmodeller.eos._saxena import H2_SF87
from atmodeller.utilities import unit_conversion


def test_Ar(check_values) -> None:
    """:cite:t:`SF87{Table 1}`"""
    model: RealGas = check_values.get_eos_model("Ar", "cs_saxena87")
    expected: float = 7.41624600755374
    check_values.compressibility(2510, 100e3, model, expected)


def test_CH4(check_values) -> None:
    """:cite:t:`SF87{Table 1}`"""
    model: RealGas = check_values.get_eos_model("CH4", "cs_shi92")
    expected: float = 17.77499804453072
    check_values.compressibility(1912, 159e3, model, expected)


def test_CO2(check_values) -> None:
    """:cite:t:`SF87{Table 1}`"""
    model: RealGas = check_values.get_eos_model("CO2", "cs_shi92")
    expected: float = 33.886349109271734
    check_values.compressibility(1167, 184e3, model, expected)


def test_H2_SF87(check_values) -> None:
    """:cite:t:`SF87{Table 1}`"""
    model: RealGas = H2_SF87
    expected: float = 4.975497264839999
    check_values.compressibility(1222, 41.66e3, model, expected)


def test_N2(check_values) -> None:
    """:cite:t:`SF87{Table 1}`"""
    model: RealGas = check_values.get_eos_model("N2", "cs_saxena87")
    expected: float = 10.293087737779091
    check_values.compressibility(1573, 75e3, model, expected)


def test_O2(check_values) -> None:
    """:cite:t:`SF87{Table 1}`"""
    model: RealGas = check_values.get_eos_model("O2", "cs_shi92")
    expected: float = 12.409268281002012
    check_values.compressibility(1823, 133e3, model, expected)


def test_H2_low_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 1}`"""
    model: RealGas = check_values.get_eos_model("H2", "shi92")
    expected: float = 7279.356114821697 * unit_conversion.cm3_to_m3
    check_values.volume(873, 10, model, expected)


def test_H2_medium_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 1}`"""
    model: RealGas = check_values.get_eos_model("H2", "shi92")
    expected: float = 164.38851468757488 * unit_conversion.cm3_to_m3
    check_values.volume(873, 500, model, expected)


def test_H2_high_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 1}`"""
    model: RealGas = check_values.get_eos_model("H2", "shi92")
    expected: float = 41.97871061892679 * unit_conversion.cm3_to_m3
    check_values.volume(1473, 4000, model, expected)


def test_H2_high_pressure2_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 1}`"""
    model: RealGas = check_values.get_eos_model("H2", "shi92")
    expected: float = 20.806595067793276 * unit_conversion.cm3_to_m3
    check_values.volume(1073, 10000, model, expected)


def test_H2_high_pressure3_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 1}`"""
    model: RealGas = check_values.get_eos_model("H2", "shi92")
    expected: float = 71.50153474005484 * unit_conversion.cm3_to_m3
    check_values.volume(673, 1000, model, expected)


def test_H2S_low_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 3}`"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    expected: float = 272.7266232763035 * unit_conversion.cm3_to_m3
    check_values.volume(673, 200, model, expected)


def test_H2S_medium_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 3}`"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    expected: float = 116.55537998390933 * unit_conversion.cm3_to_m3
    check_values.volume(1873, 2000, model, expected)


def test_SO2_low_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 2}`"""
    model: RealGas = check_values.get_eos_model("SO2", "shi92")
    expected: float = 8308.036738813245 * unit_conversion.cm3_to_m3
    check_values.volume(1073, 10, model, expected)


def test_SO2_high_pressure_SS92(check_values) -> None:
    """:cite:t:`SS92{Figure 2}`"""
    model: RealGas = check_values.get_eos_model("SO2", "shi92")
    expected: float = 70.86864302460566 * unit_conversion.cm3_to_m3
    check_values.volume(1873, 4000, model, expected)


def test_lower_extrapolation(check_values) -> None:
    """Tests the lower bound extrapolation"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    temperature: float = 1073
    pressure: float = 0.01
    expected: float = 0.01
    check_values.fugacity(temperature, pressure, model, expected)


def test_upper_extrapolation(check_values) -> None:
    """Tests the upper bound extrapolation"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    temperature: float = 1073
    pressure: float = 10e3  # Maximum pressure of the highest pressure EOS in this composite EOS
    pressure_above_max: float = 20e3  # To test pressure above the maximum
    # Expected compressibility factor at 3000 K and (the maximum calibrated pressure) 10000 bar
    expected: float = 4.472116811082408
    check_values.compressibility(temperature, pressure, model, expected)
    # The upper bound extrapolation should maintain the same compressibility for higher pressures
    # for dz/dp if reinstated at some point
    # expected: float = 8.098866090720415
    check_values.compressibility(temperature, pressure_above_max, model, expected)


def test_volume_integral_standard_pressure(check_values) -> None:
    """Tests the volume integral"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    temperature: float = 1873
    pressure: float = 1.0
    # At the standard pressure the volume integral is zero by construction
    expected: float = 0.0
    check_values.volume_integral(temperature, pressure, model, expected)


def test_volume_integral_index0(check_values) -> None:
    """Tests the volume integral for the first EOS"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    # Within ranges of experimental data, otherwise the extrapolated behaviour might be unphysical
    temperature: float = 1073
    pressure: float = 1000.0
    expected: float = 0.619184870412741
    check_values.volume_integral(temperature, pressure, model, expected)


def test_volume_integral_index1(check_values) -> None:
    """Tests the volume integral for the first and second EOS"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    temperature: float = 1073
    pressure: float = 9999
    expected: float = 1.088570827561088
    check_values.volume_integral(temperature, pressure, model, expected)


def test_volume_integral_index2(check_values) -> None:
    """Tests the volume integral in the extrapolated region above the maximum pressure"""
    model: RealGas = check_values.get_eos_model("H2S", "shi92")
    temperature: float = 1073
    pressure: float = 20000
    # for dz/dp if reinstated at some point
    # expected: float = 1.4644445135917623
    expected: float = 1.365159989472265
    check_values.volume_integral(temperature, pressure, model, expected)


# def test_volume_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = check_values.get_eos_model("CO2", "cs_shi92")
#     check_values.check_broadcasting("volume", model)


# def test_fugacity_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = check_values.get_eos_model("CO2", "cs_shi92")
#     check_values.check_broadcasting("fugacity", model)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Utilities for tests"""

import logging

# NOTE: Don't import anything from Atmodeller otherwise beartype can't wrap the tests for runtime
# type checking
import numpy as np
import numpy.typing as npt
import pytest
from jaxtyping import ArrayLike

logger: logging.Logger = logging.getLogger("atmodeller.tests")


class Helper:
    """Helper for integral tests"""

    @classmethod
    def isclose(
        cls,
        solution: dict[str, ArrayLike],
        target: dict[str, ArrayLike],
        *,
        log: bool = False,
        rtol: float = 1.0e-6,
        atol: float = 1.0e-6,
    ) -> np.bool_:
        """Determines if the solution is close to a target within tolerance.

        Only the values in `solution` and `target` that have matching keys will be compared.

        Args:
            solution: Dictionary of the solution values
            target: Dictionary of the target values
            log: Compare closeness in log-space. Defaults to False.
            rtol: Relative tolerance. Defaults to 1.0e-6.
            atol: Absolute tolerance. Defaults to 1.0e-6.

        Returns:
            True if the solution is close to the target, otherwise False
        """
        # Find the intersection of keys
        intersection_keys: set[str] = solution.keys() & target.keys()
        logger.info("Keys for comparison = %s", intersection_keys)

        # Create new dictionaries with the intersecting keys
        solution_compare: dict[str, ArrayLike] = {key: solution[key] for key in intersection_keys}
        target_compare: dict[str, ArrayLike] = {key: target[key] for key in intersection_keys}

        target_values: npt.NDArray[np.float64] = np.array(
            list(dict(sorted(target_compare.items())).values())
        )
        logger.debug("target_values = %s", target_values)
        solution_values: npt.NDArray[np.float64] = np.array(
            list(dict(sorted(solution_compare.items())).values())
        )
        logger.debug("solution_values = %s", solution_values)
        if log:
            target_values = np.log10(target_values)
            solution_values = np.log10(solution_values)

        isclose: npt.NDArray[np.bool_] = np.isclose(
            target_values, solution_values, rtol=rtol, atol=atol
        )
        logger.debug("isclose = %s", isclose)

        return isclose.all()


@pytest.fixture
def helper():
    return Helper()



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from pathlib import Path

sys.path.insert(0, os.path.abspath("."))

import sphinx.builders.latex.transforms

# Get path to directory containing this file, conf.py.
DOCS_DIRECTORY: Path = Path(__file__).resolve().parent

# -- Project information -----------------------------------------------------

project = "Atmodeller"
copyright = "2024, Dan J. Bower"
author = "Dan J. Bower"

# The full version, including alpha/beta/rc tags
release = "0.8.2"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "nbsphinx_link",
    "sphinxcontrib.bibtex",
    "sphinx_rtd_theme",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
]

bibtex_bibfiles = ["refs.bib"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = "alabaster"
html_theme = "sphinx_rtd_theme"
html_logo = "logo200x200.png"
html_context = {
    "display_github": True,
    "github_user": "ExPlanetology",
    "github_repo": "atmodeller",
    "github_version": "main",
    "conf_py_path": "/docs/",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

numfig = True

bibtex_default_style = "alpha"
bibtex_reference_style = "author_year"

pygments_style = "sphinx"

autodoc_member_order = "bysource"

autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
    "inherited-members": True,
    "undoc-members": False,
    "private-members": False,
}

latex_logo = "logo.png"
# Disable the automatic inclusion of the Python module index in the LaTeX PDF output
latex_domain_indices = False

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# https://sphinxcontrib-bibtex.readthedocs.io/en/latest/usage.html#known-issues-and-workarounds


class DummyTransform(sphinx.builders.latex.transforms.BibliographyTransform):
    def run(self, **kwargs):
        pass


sphinx.builders.latex.transforms.BibliographyTransform = DummyTransform


# https://stackoverflow.com/questions/62398231/building-docs-fails-due-to-missing-pandoc
def ensure_pandoc_installed(_):
    import pypandoc

    # Download pandoc if necessary. If pandoc is already installed and on the PATH, the installed
    # version will be used. Otherwise, we will download a copy of pandoc into docs/bin/ and add
    # that to our PATH.
    pandoc_dir: Path = DOCS_DIRECTORY / "bin"

    # Add dir containing pandoc binary to the PATH environment variable, if not already included
    if str(pandoc_dir) not in os.environ["PATH"].split(os.pathsep):
        os.environ["PATH"] += os.pathsep + str(pandoc_dir)
    pypandoc.ensure_pandoc_installed(
        targetfolder=str(pandoc_dir),
        delete_installer=True,
    )


def setup(app):
    app.connect("builder-inited", ensure_pandoc_installed)


# This avoids a warning about unable to copy the examples notebook
# Path to the existing file in the build output
destination_path = Path("_build/html/examples.ipynb")

# Remove the file if it already exists
if destination_path.exists():
    destination_path.unlink()
